from pathlib import Path
import json
import torch
import torch.nn.functional as F

from perceptor import models
from perceptor.losses.interface import LossInterface


class CLIP(LossInterface):
    def __init__(self, name="ViT-B-32", precision="fp32", jit=False):
        """
        Args:
            name: name of the clip model. Available models are:
                - RN50
                - RN101
                - RN50x4
                - RN50x16
                - RN50x64
                - ViT-B-32
                - ViT-B-16
                - ViT-L-14
                - ViT-L-14-336
        """
        super().__init__()
        self.name = name
        self.model = models.CLIP(name, precision, jit)
        self.encodings = None
        self.weights = None
        if name in ("ViT-L-14", "ViT-L-14-336"):
            self.multiplier = 0.01
        else:
            self.multiplier = 1.0

    @property
    def device(self):
        return next(iter(self.model.parameters())).device

    def mul_(self, multiplier):
        self.multiplier *= multiplier
        return self

    def add_texts_(self, texts, weights=None):
        return self.add_encodings_(self.model.encode_texts(texts), weights)

    def add_images_(self, images, weights=None):
        return self.add_encodings_(self.model.encode_images(images), weights)

    def add_text_off_(self, weight=None):
        textoff_json = json.loads(
            Path("perceptor/losses/clip/vectors/textoff.json").read_text()
        )
        if self.name in textoff_json:
            textoff = torch.tensor(textoff_json[self.name])
            return self.add_encodings_(textoff, weight)
        else:
            raise ValueError(f"There is no textoff for this model: {self.name}")

    def add_encodings_(
        self,
        encodings,
        weights=None,
    ):
        if isinstance(weights, list) or isinstance(weights, tuple):
            weights = torch.tensor(weights)
        elif weights is None:
            weights = torch.ones_like(encodings[:, 0])

        if self.encodings is None:
            self.encodings = torch.nn.Parameter(
                F.normalize(encodings).to(self.device), requires_grad=False
            )
            self.weights = torch.nn.Parameter(
                weights.to(self.device),
                requires_grad=False,
            )
        else:
            self.encodings = torch.nn.Parameter(
                torch.cat([self.encodings, F.normalize(encodings).to(self.device)]),
                requires_grad=False,
            )
            self.weights = torch.nn.Parameter(
                torch.cat([self.weights, weights.to(self.device)]),
                requires_grad=False,
            )
        return self

    def forward(self, images):
        image_encodings = self.model.encode_images(images)
        spherical_distance = (
            (image_encodings[:, None] - self.encodings[None, :])
            .norm(dim=2)
            .div(2)
            .arcsin()
            .square()
            .mul(2)
        )
        return (spherical_distance * self.weights).mean().mul(self.multiplier)


def test_clip_loss():
    loss = (
        CLIP().add_texts_(["hello", "world"]).add_images_(torch.randn(1, 3, 256, 256))
    )
    loss(torch.randn(1, 3, 256, 256))



from collections import OrderedDict
import numpy as np
import timm
import torch
from torch import nn


class LayerNorm(nn.LayerNorm):
    """Subclass torch's LayerNorm to handle fp16."""

    def forward(self, x: torch.Tensor):
        orig_type = x.dtype
        ret = super().forward(x.type(torch.float32))
        return ret.type(orig_type)


class QuickGELU(nn.Module):
    def forward(self, x: torch.Tensor):
        return x * torch.sigmoid(1.702 * x)


class ResidualAttentionBlock(nn.Module):
    def __init__(self, d_model: int, n_head: int, attn_mask: torch.Tensor = None):
        super().__init__()

        self.attn = nn.MultiheadAttention(d_model, n_head)
        self.ln_1 = LayerNorm(d_model)
        self.mlp = nn.Sequential(
            OrderedDict(
                [
                    ("c_fc", nn.Linear(d_model, d_model * 4)),
                    ("gelu", QuickGELU()),
                    ("c_proj", nn.Linear(d_model * 4, d_model)),
                ]
            )
        )
        self.ln_2 = LayerNorm(d_model)
        self.attn_mask = attn_mask

    def attention(self, x: torch.Tensor):
        self.attn_mask = (
            self.attn_mask.to(dtype=x.dtype, device=x.device)
            if self.attn_mask is not None
            else None
        )
        return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x: torch.Tensor):
        x = x + self.attention(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class Transformer(nn.Module):
    def __init__(
        self, width: int, layers: int, heads: int, attn_mask: torch.Tensor = None
    ):
        super().__init__()
        self.width = width
        self.layers = layers
        self.resblocks = nn.Sequential(
            *[ResidualAttentionBlock(width, heads, attn_mask) for _ in range(layers)]
        )

    def forward(self, x: torch.Tensor):
        return self.resblocks(x)


class CLIP(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        # vision
        vision_width: int,
        vision_model: nn.Module,
        # text
        context_length: int,
        vocab_size: int,
        transformer_width: int,
        transformer_heads: int,
        transformer_layers: int,
        **kwargs,
    ):
        super().__init__()

        self.context_length = context_length
        self.vision_width = vision_width

        self.visual = vision_model

        self.transformer = Transformer(
            width=transformer_width,
            layers=transformer_layers,
            heads=transformer_heads,
            attn_mask=self.build_attention_mask(),
        )

        self.vocab_size = vocab_size
        self.token_embedding = nn.Embedding(vocab_size, transformer_width)
        self.positional_embedding = nn.Parameter(
            torch.empty(self.context_length, transformer_width)
        )
        self.ln_final = LayerNorm(transformer_width)

        self.image_projection = nn.Parameter(torch.empty(vision_width, embed_dim))
        self.text_projection = nn.Parameter(torch.empty(transformer_width, embed_dim))
        self.logit_scale = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

        self.initialize_parameters()

    def initialize_parameters(self):
        nn.init.normal_(self.token_embedding.weight, std=0.02)
        nn.init.normal_(self.positional_embedding, std=0.01)

        proj_std = (self.transformer.width**-0.5) * (
            (2 * self.transformer.layers) ** -0.5
        )
        attn_std = self.transformer.width**-0.5
        fc_std = (2 * self.transformer.width) ** -0.5
        for block in self.transformer.resblocks:
            nn.init.normal_(block.attn.in_proj_weight, std=attn_std)
            nn.init.normal_(block.attn.out_proj.weight, std=proj_std)
            nn.init.normal_(block.mlp.c_fc.weight, std=fc_std)
            nn.init.normal_(block.mlp.c_proj.weight, std=proj_std)

        nn.init.normal_(self.image_projection, std=self.vision_width**-0.5)
        nn.init.normal_(self.text_projection, std=self.transformer.width**-0.5)

    def build_attention_mask(self):
        # lazily create causal attention mask, with full attention between the vision tokens
        # pytorch uses additive attention mask; fill with -inf
        mask = torch.empty(self.context_length, self.context_length)
        mask.fill_(float("-inf"))
        mask.triu_(1)  # zero out the lower diagonal
        return mask

    def encode_image(self, image):
        x = self.visual(image)
        x = x @ self.image_projection

        return x

    def encode_text(self, text):
        x = self.token_embedding(text)  # [batch_size, n_ctx, d_model]
        x = x + self.positional_embedding
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = self.ln_final(x)

        # x.shape = [batch_size, n_ctx, transformer.width]
        # take features from the eot embedding (eot_token is the highest number in each sequence)
        x = x[torch.arange(x.shape[0]), text.argmax(dim=-1)] @ self.text_projection

        return x

    def forward(self, image, text):
        image_embed = self.encode_image(image)
        text_embed = self.encode_text(text)

        return {
            "image_embed": image_embed,
            "text_embed": text_embed,
            "logit_scale": self.logit_scale.exp(),
        }


class SIMCLR(nn.Module):
    def __init__(
        self,
        # vision
        vision_width: int,
        vision_model: nn.Module,
        # ssl
        ssl_mlp_dim: int,
        ssl_emb_dim: int,
        **kwargs,
    ):
        super().__init__()

        self.vision_width = vision_width
        self.visual = vision_model

        self.image_mlp = self._build_mlp(
            in_dim=vision_width, mlp_dim=ssl_mlp_dim, out_dim=ssl_emb_dim
        )

    def _build_mlp(self, in_dim, mlp_dim, out_dim):
        return nn.Sequential(
            OrderedDict(
                [
                    ("layer1", nn.Linear(in_dim, mlp_dim)),
                    ("bn1", nn.SyncBatchNorm(mlp_dim)),
                    ("relu1", nn.ReLU(inplace=True)),
                    ("layer2", nn.Linear(mlp_dim, mlp_dim)),
                    ("bn2", nn.SyncBatchNorm(mlp_dim)),
                    ("relu2", nn.ReLU(inplace=True)),
                    ("layer3", nn.Linear(mlp_dim, out_dim)),
                ]
            )
        )

    def encode_image(self, image):
        x = self.visual(image)

        return x

    def forward(self, aug1, aug2):
        h1 = self.visual(aug1)
        h2 = self.visual(aug2)

        aug1_embed = self.image_mlp(h1)
        aug2_embed = self.image_mlp(h2)

        return {"aug1_embed": aug1_embed, "aug2_embed": aug2_embed}


class SLIP(CLIP):
    def __init__(
        self,
        ssl_mlp_dim: int,
        ssl_emb_dim: int,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.image_mlp = self._build_mlp(
            in_dim=self.vision_width, mlp_dim=ssl_mlp_dim, out_dim=ssl_emb_dim
        )

    def _build_mlp(self, in_dim, mlp_dim, out_dim):
        return nn.Sequential(
            OrderedDict(
                [
                    ("layer1", nn.Linear(in_dim, mlp_dim)),
                    ("bn1", nn.SyncBatchNorm(mlp_dim)),
                    ("relu1", nn.ReLU(inplace=True)),
                    ("layer2", nn.Linear(mlp_dim, mlp_dim)),
                    ("bn2", nn.SyncBatchNorm(mlp_dim)),
                    ("relu2", nn.ReLU(inplace=True)),
                    ("layer3", nn.Linear(mlp_dim, out_dim)),
                ]
            )
        )

    def forward(self, image, text, aug1, aug2):
        aug1_embed = self.image_mlp(self.visual(aug1))
        aug2_embed = self.image_mlp(self.visual(aug2))

        image_embed = self.encode_image(image)
        text_embed = self.encode_text(text)

        return {
            "image_embed": image_embed,
            "text_embed": text_embed,
            "logit_scale": self.logit_scale.exp(),
            "aug1_embed": aug1_embed,
            "aug2_embed": aug2_embed,
        }


def get_loss(model, ssl_temp, ssl_scale):
    if model.startswith("SLIP"):
        ssl_loss = losses.SIMCLRLoss(temperature=ssl_temp)
        return losses.SLIPLoss(ssl_loss, ssl_scale)
    if model.startswith("CLIP"):
        return losses.CLIPLoss()
    if model.startswith("SIMCLR"):
        return losses.SIMCLRLoss(temperature=ssl_temp)


def get_metric_names(model):
    if model.startswith("SLIP"):
        return ["loss", "clip_loss", "ssl_loss", "clip_acc", "ssl_acc"]
    elif model.startswith("CLIP"):
        return ["loss", "clip_loss", "clip_acc"]
    else:
        return ["loss", "ssl_loss", "ssl_acc"]


@timm.models.registry.register_model
def vit_small_mocov3_patch16_224(**kwargs):
    model_kwargs = dict(patch_size=16, embed_dim=384, depth=12, num_heads=12, **kwargs)
    model = timm.models.vision_transformer._create_vision_transformer(
        "vit_small_patch16_224", **model_kwargs
    )

    return model


def CLIP_VITS16(**kwargs):
    vision_model = timm.create_model("vit_small_mocov3_patch16_224", num_classes=0)
    model = CLIP(
        embed_dim=512,
        vision_width=384,
        vision_model=vision_model,
        context_length=77,
        vocab_size=49408,
        transformer_width=512,
        transformer_heads=8,
        transformer_layers=12,
        **kwargs,
    )

    return model


def SIMCLR_VITS16(**kwargs):
    vision_model = timm.create_model("vit_small_mocov3_patch16_224", num_classes=0)
    model = SIMCLR(vision_width=384, vision_model=vision_model, **kwargs)

    return model


def SLIP_VITS16(**kwargs):
    vision_model = timm.create_model("vit_small_mocov3_patch16_224", num_classes=0)
    model = SLIP(
        embed_dim=512,
        vision_width=384,
        vision_model=vision_model,
        context_length=77,
        vocab_size=49408,
        transformer_width=512,
        transformer_heads=8,
        transformer_layers=12,
        **kwargs,
    )

    return model


def CLIP_VITB16(**kwargs):
    vision_model = timm.create_model("vit_base_patch16_224", num_classes=0)
    model = CLIP(
        embed_dim=512,
        vision_width=768,
        vision_model=vision_model,
        context_length=77,
        vocab_size=49408,
        transformer_width=512,
        transformer_heads=8,
        transformer_layers=12,
        **kwargs,
    )

    return model


def SIMCLR_VITB16(**kwargs):
    vision_model = timm.create_model("vit_base_patch16_224", num_classes=0)
    model = SIMCLR(vision_width=768, vision_model=vision_model, **kwargs)

    return model


def SLIP_VITB16(**kwargs):
    vision_model = timm.create_model("vit_base_patch16_224", num_classes=0)
    model = SLIP(
        embed_dim=512,
        vision_width=768,
        vision_model=vision_model,
        context_length=77,
        vocab_size=49408,
        transformer_width=512,
        transformer_heads=8,
        transformer_layers=12,
        **kwargs,
    )

    return model


def CLIP_VITL16(**kwargs):
    vision_model = timm.create_model("vit_large_patch16_224", num_classes=0)
    model = CLIP(
        embed_dim=512,
        vision_width=1024,
        vision_model=vision_model,
        context_length=77,
        vocab_size=49408,
        transformer_width=512,
        transformer_heads=8,
        transformer_layers=12,
        **kwargs,
    )

    return model


def SIMCLR_VITL16(**kwargs):
    vision_model = timm.create_model("vit_large_patch16_224", num_classes=0)
    model = SIMCLR(vision_width=1024, vision_model=vision_model, **kwargs)

    return model


def SLIP_VITL16(**kwargs):
    vision_model = timm.create_model("vit_large_patch16_224", num_classes=0)
    model = SLIP(
        embed_dim=512,
        vision_width=1024,
        vision_model=vision_model,
        context_length=77,
        vocab_size=49408,
        transformer_width=512,
        transformer_heads=8,
        transformer_layers=12,
        **kwargs,
    )

    return model



import torch
import torch.nn.functional as F
from basicsr.utils.download_util import load_file_from_url

from .model_creation import create_clip_model
from perceptor import utils


CHECKPOINT_URLS = {
    "clip/image-enc": "https://openaipublic.blob.core.windows.net/diffusion/dec-2021/clip_image_enc.pt",
    "clip/text-enc": "https://openaipublic.blob.core.windows.net/diffusion/dec-2021/clip_text_enc.pt",
}


@utils.cache
class GlideCLIP(torch.nn.Module):
    def __init__(self):
        """CLIP model trained to handle noisy images"""
        super().__init__()

        start_device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )

        self.model = create_clip_model(device=start_device).eval().requires_grad_(False)
        self.model.image_encoder.load_state_dict(
            torch.load(
                load_file_from_url(CHECKPOINT_URLS["clip/image-enc"], "models"),
                map_location=start_device,
            )
        )
        self.model.text_encoder.load_state_dict(
            torch.load(
                load_file_from_url(CHECKPOINT_URLS["clip/text-enc"], "models"),
                map_location=start_device,
            )
        )

    @property
    def device(self):
        return next(iter(self.parameters())).device

    def encode_texts(self, text_prompts):
        return F.normalize(self.model.text_embeddings(text_prompts))

    def encode_images(self, diffused, ts):
        """
        Args:
            diffused: [batch_size, 3, 64, 64] diffused images between 0 and 1
            t: [batch_size] timestep between 0 and 999 where 0 is without noise.
        """
        return F.normalize(
            self.model.image_embeddings(
                diffused.to(self.device).mul(2).sub(1),
                ts.to(self.device),
            )
        )

    def forward(self, _):
        raise NotImplementedError



"""
MiDaS model from https://github.com/isl-org/MiDaS and

Inference impementation from https://github.com/alembics/disco-diffusion/blob/main/disco.py
"""
from lantern import Tensor
import torch
from torch import nn
import torchvision.transforms as T
from basicsr.utils.download_util import load_file_from_url

from perceptor import utils, transforms
from .dpt_depth import DPTDepthModel
from .midas_net import MidasNet
from .midas_net_custom import MidasNet_small

# from .transforms import Resize, NormalizeImage, PrepareForNet

checkpoints = {
    "midas_v21_small": "https://github.com/AlexeyAB/MiDaS/releases/download/midas_dpt/midas_v21_small-70d6b9c8.pt",
    "midas_v21": "https://github.com/AlexeyAB/MiDaS/releases/download/midas_dpt/midas_v21-f6b98070.pt",
    "dpt_large": "https://github.com/intel-isl/DPT/releases/download/1_0/dpt_large-midas-2f21e586.pt",
    "dpt_hybrid": "https://github.com/intel-isl/DPT/releases/download/1_0/dpt_hybrid-midas-501f0c75.pt",
    "dpt_hybrid_nyu": "https://github.com/intel-isl/DPT/releases/download/1_0/dpt_hybrid_nyu-2ce69ec7.pt",
    "dpt_hybrid_kitti": "https://github.com/intel-isl/DPT/releases/download/1_0/dpt_hybrid_kitti-cb926ef4.pt",
}


@utils.cache
class MidasDepth(nn.Module):
    def __init__(self, name="dpt_large", optimize=True):
        """
        Args:
            name (str, optional): Name of the model. Suggested value:
                dpt_large - highest quality
                dpt_hybrid - moderately less quality, but better speed on CPU and slower GPUs
                midas_v21_small - real-time applications on resource-constrained devices
        """
        super().__init__()
        self.name = name
        self.model = None
        self.optimize = optimize

        checkpoint_path = load_file_from_url(
            checkpoints[self.name],
            "models",
        )

        if self.name == "dpt_large":  # DPT-Large
            self.model = DPTDepthModel(
                path=checkpoint_path,
                backbone="vitl16_384",
                non_negative=True,
            )
            self.net_w, self.net_h = 384, 384
            self.resize_mode = "minimal"
            self.normalization = T.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        elif self.name == "dpt_hybrid":  # DPT-Hybrid
            self.model = DPTDepthModel(
                path=checkpoint_path,
                backbone="vitb_rn50_384",
                non_negative=True,
            )
            self.net_w, self.net_h = 384, 384
            self.resize_mode = "minimal"
            self.normalization = T.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        elif self.name == "dpt_hybrid_nyu":  # DPT-Hybrid-NYU
            self.model = DPTDepthModel(
                path=checkpoint_path,
                backbone="vitb_rn50_384",
                non_negative=True,
            )
            self.net_w, self.net_h = 384, 384
            self.resize_mode = "minimal"
            self.normalization = T.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        elif self.name == "dpt_hybrid_kitti":  # Assumed same as DPT-Hybrid-NYU
            self.model = DPTDepthModel(
                path=checkpoint_path,
                backbone="vitb_rn50_384",
                non_negative=True,
            )
            self.net_w, self.net_h = 384, 384
            self.resize_mode = "minimal"
            self.normalization = T.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        elif self.name == "midas_v21":
            self.model = MidasNet(checkpoint_path, non_negative=True)
            self.net_w, self.net_h = 384, 384
            self.resize_mode = "upper_bound"
            self.normalization = T.Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            )
        elif self.name == "midas_v21_small":
            self.model = MidasNet_small(
                checkpoint_path,
                features=64,
                backbone="efficientnet_lite3",
                exportable=True,
                non_negative=True,
                blocks={"expand": True},
            )
            self.net_w, self.net_h = 256, 256
            self.resize_mode = "upper_bound"
            self.normalization = T.Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            )
        else:
            raise ValueError(f"midas_model_type '{self.name}' not implemented")

        self.image_size = (self.net_h, self.net_w)

        self.model.eval()
        self.model.requires_grad_(False)

    def to(self, device):
        if self.optimize:
            if device == torch.device("cuda"):
                self.model = self.model.to(memory_format=torch.channels_last)
                self.model = self.model.half()
        return super().to(device)

    @torch.cuda.amp.autocast()
    def forward(self, images: Tensor.dims("NCHW")) -> Tensor.dims("NCHW"):
        if images.shape[-2:] != self.image_size:
            images = transforms.resize(
                images,
                out_shape=self.image_size,
            )
        return -self.model(self.normalization(images)).float()[:, None]


def test_midas_depth():
    import requests
    from PIL import Image
    import torch
    import torchvision.transforms.functional as TF

    image_url = "http://images.cocodataset.org/val2017/000000039769.jpg"
    images = (
        TF.to_tensor(Image.open(requests.get(image_url, stream=True).raw))[None]
        .cuda()
        .requires_grad_()
    )

    model = MidasDepth().cuda()

    with torch.enable_grad():
        depths = model(images)
        assert len(depths.shape) == 4
        depths.mean().backward()

    assert images.grad is not None

    depths = transforms.resize(depths, out_shape=images.shape[-2:])

    utils.pil_image(
        torch.cat(
            [
                images,
                (depths.repeat(1, 3, 1, 1) - depths.min())
                / (depths.max() - depths.min()),
            ]
        )
    ).save("tests/midas_depth.png")



# https://github.com/mlomnitz/DiffJPEG
import itertools
import numpy as np
import torch
from torch import nn
from .utils import c_table, y_table, diff_round


class rgb_to_ycbcr_jpeg(nn.Module):
    """Converts RGB image to YCbCr
    Input:
        image(tensor): batch x 3 x height x width
    Outpput:
        result(tensor): batch x height x width x 3
    """

    def __init__(self):
        super(rgb_to_ycbcr_jpeg, self).__init__()
        matrix = np.array(
            [
                [0.299, 0.587, 0.114],
                [-0.168736, -0.331264, 0.5],
                [0.5, -0.418688, -0.081312],
            ],
            dtype=np.float32,
        ).T
        self.shift = nn.Parameter(
            torch.tensor([0.0, 128.0, 128.0]), requires_grad=False
        )
        #
        self.matrix = nn.Parameter(torch.from_numpy(matrix), requires_grad=False)

    def forward(self, image):
        image = image.permute(0, 2, 3, 1)
        result = torch.tensordot(image, self.matrix, dims=1) + self.shift
        #    result = torch.from_numpy(result)
        result.view(image.shape)
        return result


class chroma_subsampling(nn.Module):
    """Chroma subsampling on CbCv channels
    Input:
        image(tensor): batch x height x width x 3
    Output:
        y(tensor): batch x height x width
        cb(tensor): batch x height/2 x width/2
        cr(tensor): batch x height/2 x width/2
    """

    def __init__(self):
        super(chroma_subsampling, self).__init__()

    def forward(self, image):
        image_2 = image.permute(0, 3, 1, 2).clone()
        avg_pool = nn.AvgPool2d(kernel_size=2, stride=(2, 2), count_include_pad=False)
        cb = avg_pool(image_2[:, 1, :, :].unsqueeze(1))
        cr = avg_pool(image_2[:, 2, :, :].unsqueeze(1))
        cb = cb.permute(0, 2, 3, 1)
        cr = cr.permute(0, 2, 3, 1)
        return image[:, :, :, 0], cb.squeeze(3), cr.squeeze(3)


class block_splitting(nn.Module):
    """Splitting image into patches
    Input:
        image(tensor): batch x height x width
    Output:
        patch(tensor):  batch x h*w/64 x h x w
    """

    def __init__(self):
        super(block_splitting, self).__init__()
        self.k = 8

    def forward(self, image):
        height, width = image.shape[1:3]
        batch_size = image.shape[0]
        image_reshaped = image.view(batch_size, height // self.k, self.k, -1, self.k)
        image_transposed = image_reshaped.permute(0, 1, 3, 2, 4)
        return image_transposed.contiguous().view(batch_size, -1, self.k, self.k)


class dct_8x8(nn.Module):
    """Discrete Cosine Transformation
    Input:
        image(tensor): batch x height x width
    Output:
        dcp(tensor): batch x height x width
    """

    def __init__(self):
        super(dct_8x8, self).__init__()
        tensor = np.zeros((8, 8, 8, 8), dtype=np.float32)
        for x, y, u, v in itertools.product(range(8), repeat=4):
            tensor[x, y, u, v] = np.cos((2 * x + 1) * u * np.pi / 16) * np.cos(
                (2 * y + 1) * v * np.pi / 16
            )
        alpha = np.array([1.0 / np.sqrt(2)] + [1] * 7)

        self.tensor = nn.Parameter(
            torch.from_numpy(tensor).float(), requires_grad=False
        )
        self.scale = nn.Parameter(
            torch.from_numpy(np.outer(alpha, alpha) * 0.25).float(), requires_grad=False
        )

    def forward(self, image):
        image = image - 128
        result = self.scale * torch.tensordot(image, self.tensor, dims=2)
        result.view(image.shape)
        return result


class y_quantize(nn.Module):
    """JPEG Quantization for Y channel
    Input:
        image(tensor): batch x height x width
        rounding(function): rounding function to use
        factor(float): Degree of compression
    Output:
        image(tensor): batch x height x width
    """

    def __init__(self, rounding, factor=1):
        super(y_quantize, self).__init__()
        self.rounding = rounding
        self.factor = factor
        self.y_table = nn.Parameter(y_table(), requires_grad=False)

    def forward(self, image):
        image = image.float() / (self.y_table * self.factor)
        image = self.rounding(image)
        return image


class c_quantize(nn.Module):
    """JPEG Quantization for CrCb channels
    Input:
        image(tensor): batch x height x width
        rounding(function): rounding function to use
        factor(float): Degree of compression
    Output:
        image(tensor): batch x height x width
    """

    def __init__(self, rounding, factor=1):
        super(c_quantize, self).__init__()
        self.rounding = rounding
        self.factor = factor
        self.c_table = nn.Parameter(c_table(), requires_grad=False)

    def forward(self, image):
        image = image.float() / (self.c_table * self.factor)
        image = self.rounding(image)
        return image


class compress_jpeg(nn.Module):
    """Full JPEG compression algortihm
    Input:
        imgs(tensor): batch x 3 x height x width
        rounding(function): rounding function to use
        factor(float): Compression factor
    Ouput:
        compressed(dict(tensor)): batch x h*w/64 x 8 x 8
    """

    def __init__(self, rounding=diff_round, factor=1):
        super(compress_jpeg, self).__init__()
        self.l1 = nn.Sequential(rgb_to_ycbcr_jpeg(), chroma_subsampling())
        self.l2 = nn.Sequential(block_splitting(), dct_8x8())
        self.c_quantize = c_quantize(rounding=rounding, factor=factor)
        self.y_quantize = y_quantize(rounding=rounding, factor=factor)

    def forward(self, image):
        y, cb, cr = self.l1(image * 255)
        components = {"y": y, "cb": cb, "cr": cr}
        for k in components.keys():
            comp = self.l2(components[k])
            if k in ("cb", "cr"):
                comp = self.c_quantize(comp)
            else:
                comp = self.y_quantize(comp)

            components[k] = comp

        return components["y"], components["cb"], components["cr"]



import torch
import torch.nn as nn
from .common import *
from .updown import Upsample2d


def skip(
    num_input_channels=2,
    num_output_channels=3,
    num_channels_down=[16, 32, 64, 128, 128],
    num_channels_up=[16, 32, 64, 128, 128],
    num_channels_skip=[4, 4, 4, 4, 4],
    filter_size_down=3,
    filter_size_up=3,
    filter_skip_size=1,
    need_sigmoid=True,
    need_bias=True,
    decorr_rgb=False,
    pad="zero",
    upsample_mode="nearest",
    downsample_mode="stride",
    act_fun="LeakyReLU",
    need1x1_up=True,
    offset_groups=0,
    offset_type="full",
):
    """Assembles encoder-decoder with skip connections.
    Arguments:
        act_fun: Either string 'LeakyReLU|Swish|ELU|none' or module (e.g. nn.ReLU)
        pad (string): zero|reflection (default: 'zero')
        upsample_mode (string): 'nearest|bilinear|cubic|lanczos3' (default: 'nearest')
        downsample_mode (string): 'stride|avg|max|linear|cubic|lanczos3' (default: 'stride')
    """
    assert len(num_channels_down) == len(num_channels_up) == len(num_channels_skip)

    n_scales = len(num_channels_down)

    if not (isinstance(upsample_mode, list) or isinstance(upsample_mode, tuple)):
        upsample_mode = [upsample_mode] * n_scales

    if not (isinstance(downsample_mode, list) or isinstance(downsample_mode, tuple)):
        downsample_mode = [downsample_mode] * n_scales

    if not (isinstance(filter_size_down, list) or isinstance(filter_size_down, tuple)):
        filter_size_down = [filter_size_down] * n_scales

    if not (isinstance(filter_size_up, list) or isinstance(filter_size_up, tuple)):
        filter_size_up = [filter_size_up] * n_scales

    if not (isinstance(offset_groups, list) or isinstance(offset_groups, tuple)):
        offset_groups = [offset_groups] * n_scales

    last_scale = n_scales - 1

    cur_depth = None

    model = nn.Sequential()
    model_tmp = model

    input_depth = num_input_channels
    for i in range(len(num_channels_down)):

        deeper = nn.Sequential()
        skip = nn.Sequential()

        if num_channels_skip[i] != 0:
            model_tmp.add(Concat(1, skip, deeper))
        else:
            model_tmp.add(deeper)

        model_tmp.add(
            bn(
                num_channels_skip[i]
                + (num_channels_up[i + 1] if i < last_scale else num_channels_down[i])
            )
        )

        if num_channels_skip[i] != 0:
            skip.add(
                conv(
                    input_depth,
                    num_channels_skip[i],
                    filter_skip_size,
                    bias=need_bias,
                    pad=pad,
                    offset_groups=offset_groups[i],
                    offset_type=offset_type,
                )
            )
            skip.add(bn(num_channels_skip[i]))
            skip.add(act(act_fun))

        # skip.add(Concat(2, GenNoise(nums_noise[i]), skip_part))

        deeper.add(
            conv(
                input_depth,
                num_channels_down[i],
                filter_size_down[i],
                2,
                bias=need_bias,
                pad=pad,
                downsample_mode=downsample_mode[i],
                offset_groups=offset_groups[i],
                offset_type=offset_type,
            )
        )
        deeper.add(bn(num_channels_down[i]))
        deeper.add(act(act_fun))

        deeper.add(
            conv(
                num_channels_down[i],
                num_channels_down[i],
                filter_size_down[i],
                bias=need_bias,
                pad=pad,
                offset_groups=offset_groups[i],
                offset_type=offset_type,
            )
        )
        deeper.add(bn(num_channels_down[i]))
        deeper.add(act(act_fun))

        deeper_main = nn.Sequential()

        if i == len(num_channels_down) - 1:
            # The deepest
            k = num_channels_down[i]
        else:
            deeper.add(deeper_main)
            k = num_channels_up[i + 1]

        deeper.add(up(upsample_mode[i]))

        model_tmp.add(
            conv(
                num_channels_skip[i] + k,
                num_channels_up[i],
                filter_size_up[i],
                1,
                bias=need_bias,
                pad=pad,
                offset_groups=offset_groups[i],
                offset_type=offset_type,
            )
        )
        model_tmp.add(bn(num_channels_up[i]))
        model_tmp.add(act(act_fun))

        if need1x1_up:
            model_tmp.add(
                conv(num_channels_up[i], num_channels_up[i], 1, bias=need_bias, pad=pad)
            )
            model_tmp.add(bn(num_channels_up[i]))
            model_tmp.add(act(act_fun))

        input_depth = num_channels_down[i]
        model_tmp = deeper_main

    model.add(conv(num_channels_up[0], num_output_channels, 1, bias=need_bias, pad=pad))
    if decorr_rgb:
        model.add(DecorrelatedColorsToRGB())
    if need_sigmoid:
        model.add(nn.Sigmoid())

    return model



from . import cc12m_1, yfcc_1, yfcc_2, wikiart_256


MODELS = {
    "cc12m_1": cc12m_1.CC12M1Model,
    "cc12m_1_cfg": cc12m_1.CC12M1Model,
    "yfcc_1": yfcc_1.YFCC1Model,
    "yfcc_2": yfcc_2.YFCC2Model,
    "wikiart": wikiart_256.WikiArt256Model,
}


def get_model(model):
    return MODELS[model]


def get_models():
    return list(MODELS.keys())



import torch
import torch.nn.functional as F
from basicsr.archs.rrdbnet_arch import RRDBNet
from basicsr.utils.download_util import load_file_from_url
from huggingface_hub import hf_hub_url, cached_download

from perceptor import utils
from .real_esrganer import RealESRGANer
from .srvgg_net_compact import SRVGGNetCompact
from .custom_rrdbnet_arch import CustomRRDBNet


checkpoints = {
    "RealESRGAN_x4plus": "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.1.0/RealESRGAN_x4plus.pth",
    "RealESRGAN_x2plus": "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.1/RealESRGAN_x2plus.pth",
    "RealESRNet_x4plus": "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.1.1/RealESRNet_x4plus.pth",
    "RealESRGAN_x4plus_anime_6B": "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.2.4/RealESRGAN_x4plus_anime_6B.pth",
    "RealESRGANv2-animevideo-xsx2": "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.3.0/RealESRGANv2-animevideo-xsx2.pth",
    "RealESRGANv2-animevideo-xsx4": "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.3.0/RealESRGANv2-animevideo-xsx4.pth",
}

sberbank_checkpoints = {
    "x2": dict(
        scale=2,
        repo_id="shonenkov/rudalle-utils",
        filename="RealESRGAN_x2.pth",
    ),
    "x4": dict(
        scale=4,
        repo_id="shonenkov/rudalle-utils",
        filename="RealESRGAN_x4.pth",
    ),
    "x8": dict(
        scale=8,
        repo_id="shonenkov/rudalle-utils",
        filename="RealESRGAN_x8.pth",
    ),
}


@utils.cache
class SuperResolution(torch.nn.Module):
    def __init__(self, name="x4", half=False):
        super().__init__()
        self.name = name

        if self.name in sberbank_checkpoints:
            config = sberbank_checkpoints[name]
            self.scale = config["scale"]
            cache_dir = "models"
            config_file_url = hf_hub_url(
                repo_id=config["repo_id"], filename=config["filename"]
            )
            cached_download(
                config_file_url, cache_dir=cache_dir, force_filename=config["filename"]
            )
            model = CustomRRDBNet(
                num_in_ch=3,
                num_out_ch=3,
                num_feat=64,
                num_block=23,
                num_grow_ch=32,
                scale=self.scale,
            )
            checkpoint_path = f"{cache_dir}/{config['filename']}"
        else:
            checkpoint_path = load_file_from_url(
                checkpoints[self.name],
                "models",
            )

            if self.name in ["RealESRGAN_x4plus", "RealESRNet_x4plus"]:
                model = RRDBNet(
                    num_in_ch=3,
                    num_out_ch=3,
                    num_feat=64,
                    num_block=23,
                    num_grow_ch=32,
                    scale=4,
                )
                self.scale = 4
            elif self.name in ["RealESRGAN_x4plus_anime_6B"]:
                model = RRDBNet(
                    num_in_ch=3,
                    num_out_ch=3,
                    num_feat=64,
                    num_block=6,
                    num_grow_ch=32,
                    scale=4,
                )
                self.scale = 4
            elif self.name in ["RealESRGAN_x2plus"]:  # x2 RRDBNet model
                model = RRDBNet(
                    num_in_ch=3,
                    num_out_ch=3,
                    num_feat=64,
                    num_block=23,
                    num_grow_ch=32,
                    scale=2,
                )
                self.scale = 2
            elif self.name in [
                "RealESRGANv2-anime-xsx2",
                "RealESRGANv2-animevideo-xsx2-nousm",
                "RealESRGANv2-animevideo-xsx2",
            ]:
                model = SRVGGNetCompact(
                    num_in_ch=3,
                    num_out_ch=3,
                    num_feat=64,
                    num_conv=16,
                    upscale=2,
                    act_type="prelu",
                )
                self.scale = 2
            elif self.name in [
                "RealESRGANv2-anime-xsx4",
                "RealESRGANv2-animevideo-xsx4-nousm",
                "RealESRGANv2-animevideo-xsx4",
            ]:
                model = SRVGGNetCompact(
                    num_in_ch=3,
                    num_out_ch=3,
                    num_feat=64,
                    num_conv=16,
                    upscale=4,
                    act_type="prelu",
                )
                self.scale = 4

        model.eval()
        model.requires_grad_(False)

        self.upsampler = RealESRGANer(
            scale=self.scale,
            model_path=checkpoint_path,
            model=model,
            tile=0,
            tile_pad=10,
            pre_pad=0,
            half=half,
        )

    @torch.cuda.amp.autocast()
    def forward(self, image):
        return self.upsample(image)

    def upsample(self, image):
        return self.upsampler.enhance(
            image,
            outscale=self.scale,
        ).float()

    def downsample(self, upsampled_image, size=None):
        if size is None:
            size = (torch.tensor(upsampled_image.shape[-2:]) // self.scale).tolist()
        return F.interpolate(
            upsampled_image,
            size=size,
            mode="bilinear",
            align_corners=False,
        )



import numpy as np
import torch


def interpolant(t):
    return t * t * t * (t * (t * 6 - 15) + 10)


def generate_perlin_noise_2d(
    shape, res, tileable=(False, False), interpolant=interpolant
):
    """Generate a 2D numpy array of perlin noise.
    Args:
        shape: The shape of the generated array (tuple of two ints).
            This must be a multple of res.
        res: The number of periods of noise to generate along each
            axis (tuple of two ints). Note shape must be a multiple of
            res.
        tileable: If the noise should be tileable along each axis
            (tuple of two bools). Defaults to (False, False).
        interpolant: The interpolation function, defaults to
            t*t*t*(t*(t*6 - 15) + 10).
    Returns:
        A numpy array of shape shape with the generated noise.
    Raises:
        ValueError: If shape is not a multiple of res.
    """
    delta = (res[0] / shape[0], res[1] / shape[1])
    d = (shape[0] // res[0], shape[1] // res[1])
    grid = np.mgrid[0 : res[0] : delta[0], 0 : res[1] : delta[1]].transpose(1, 2, 0) % 1
    # Gradients
    angles = 2 * np.pi * np.random.rand(res[0] + 1, res[1] + 1)
    gradients = np.dstack((np.cos(angles), np.sin(angles)))
    if tileable[0]:
        gradients[-1, :] = gradients[0, :]
    if tileable[1]:
        gradients[:, -1] = gradients[:, 0]
    gradients = gradients.repeat(d[0], 0).repeat(d[1], 1)
    g00 = gradients[: -d[0], : -d[1]]
    g10 = gradients[d[0] :, : -d[1]]
    g01 = gradients[: -d[0], d[1] :]
    g11 = gradients[d[0] :, d[1] :]
    # Ramps
    n00 = np.sum(np.dstack((grid[:, :, 0], grid[:, :, 1])) * g00, 2)
    n10 = np.sum(np.dstack((grid[:, :, 0] - 1, grid[:, :, 1])) * g10, 2)
    n01 = np.sum(np.dstack((grid[:, :, 0], grid[:, :, 1] - 1)) * g01, 2)
    n11 = np.sum(np.dstack((grid[:, :, 0] - 1, grid[:, :, 1] - 1)) * g11, 2)
    # Interpolation
    t = interpolant(grid)
    n0 = n00 * (1 - t[:, :, 0]) + t[:, :, 0] * n10
    n1 = n01 * (1 - t[:, :, 0]) + t[:, :, 0] * n11
    return np.sqrt(2) * ((1 - t[:, :, 1]) * n0 + t[:, :, 1] * n1)


def generate_fractal_noise_2d(
    shape,
    res,
    octaves=1,
    persistence=0.5,
    lacunarity=2,
    tileable=(False, False),
    interpolant=interpolant,
):
    """Generate a 2D numpy array of fractal noise.
    Args:
        shape: The shape of the generated array (tuple of two ints).
            This must be a multiple of lacunarity**(octaves-1)*res.
        res: The number of periods of noise to generate along each
            axis (tuple of two ints). Note shape must be a multiple of
            (lacunarity**(octaves-1)*res).
        octaves: The number of octaves in the noise. Defaults to 1.
        persistence: The scaling factor between two octaves.
        lacunarity: The frequency factor between two octaves.
        tileable: If the noise should be tileable along each axis
            (tuple of two bools). Defaults to (False, False).
        interpolant: The, interpolation function, defaults to
            t*t*t*(t*(t*6 - 15) + 10).
    Returns:
        A numpy array of fractal noise and of shape shape generated by
        combining several octaves of perlin noise.
    Raises:
        ValueError: If shape is not a multiple of
            (lacunarity**(octaves-1)*res).
    """
    noise = np.zeros(shape)
    frequency = 1
    amplitude = 1
    for _ in range(octaves):
        noise += amplitude * generate_perlin_noise_2d(
            shape, (frequency * res[0], frequency * res[1]), tileable, interpolant
        )
        frequency *= lacunarity
        amplitude *= persistence
    return noise


def normalize(data):
    return (data - np.min(data)) / (np.max(data) - np.min(data))


# https://stats.stackexchange.com/a/289477
def contrast_noise(n):
    n = 0.9998 * n + 0.0001
    n1 = n / (1 - n)
    n2 = np.power(n1, -2)
    n3 = 1 / (1 + n2)
    return n3


def fractal(shape):
    n, c, h, w = shape
    # scale up roughly as power of 2
    if w > 1024 or h > 1024:
        side, octp = 2048, 6
    elif w > 512 or h > 512:
        side, octp = 1024, 5
    elif w > 256 or h > 256:
        side, octp = 512, 4
    else:
        side, octp = 256, 3

    return torch.from_numpy(
        np.stack(
            [
                np.stack(
                    [
                        contrast_noise(
                            normalize(
                                generate_fractal_noise_2d((side, side), (32, 32), octp)
                            )
                        )[:h, :w]
                        for _ in range(c)
                    ]
                )
                for _ in range(n)
            ]
        )
    ).float()


def test_fractal():
    fractal((1, 3, 256, 256))



"""
https://github.com/crowsonkb/simulacra-aesthetic-models/blob/master/LICENSE
"""

import torch
from torch import nn
from torch.nn import functional as F
from basicsr.utils.download_util import load_file_from_url

from perceptor import models

CHECKPOINT_URLS = {
    "ViT-B-32": "https://raw.githubusercontent.com/crowsonkb/simulacra-aesthetic-models/master/models/sac_public_2022_06_29_vit_b_32_linear.pth",
    "ViT-B-16": "https://raw.githubusercontent.com/crowsonkb/simulacra-aesthetic-models/master/models/sac_public_2022_06_29_vit_b_16_linear.pth",
    "ViT-L-14": "https://raw.githubusercontent.com/crowsonkb/simulacra-aesthetic-models/master/models/sac_public_2022_06_29_vit_l_14_linear.pth",
    "RN50": "https://raw.githubusercontent.com/samedii/perceptor/master/perceptor/models/simulacra_aesthetic/weights/RN50.pth",
    "RN101": "https://raw.githubusercontent.com/samedii/perceptor/master/perceptor/models/simulacra_aesthetic/weights/RN101.pth",
    "RN50x4": "https://raw.githubusercontent.com/samedii/perceptor/master/perceptor/models/simulacra_aesthetic/weights/RN50x4.pth",
    "RN50x16": "https://raw.githubusercontent.com/samedii/perceptor/master/perceptor/models/simulacra_aesthetic/weights/RN50x16.pth",
    "RN50x64": "https://raw.githubusercontent.com/samedii/perceptor/master/perceptor/models/simulacra_aesthetic/weights/RN50x64.pth",
    "ViT-L-14-336": "https://raw.githubusercontent.com/samedii/perceptor/master/perceptor/models/simulacra_aesthetic/weights/ViT-L-14-336px.pth",
}


class SimulacraAesthetic(nn.Module):
    def __init__(self, model_name="ViT-B-32"):
        """
        Simulacra aesthetic loss based on clip linear regression probe that predicts the aesthetic rating of an image.

        Args:
            model_name (str): Name of CLIP model. Available models:
                - ViT-B-32
                - ViT-B-16
                - ViT-L-14
                - RN50
                - RN101
                - RN50x4
                - RN50x16
                - RN50x64
                - ViT-L-14-336
        """
        super().__init__()

        clip_model = models.CLIP(model_name)

        checkpoint_path = load_file_from_url(
            CHECKPOINT_URLS[model_name],
            "models",
        )
        state_dict = torch.load(checkpoint_path, map_location="cpu")
        self.linear = nn.Linear(state_dict["linear.weight"].shape[1], 1)
        self.load_state_dict(state_dict)
        self.linear.eval()
        self.linear.requires_grad_(False)

        self.clip_model = clip_model

    def forward(self, images):
        encodings = self.clip_model.encode_images(images)
        return self.linear(F.normalize(encodings, dim=-1) * encodings.shape[-1] ** 0.5)


def test_simulacra_aesthetic():
    model = SimulacraAesthetic().cuda()
    model(torch.randn(1, 3, 256, 256).cuda())



import torch
import torch.nn as nn

from .layers import PatchTransformerEncoder, PixelWiseDotProduct


class mViT(nn.Module):
    def __init__(
        self,
        in_channels,
        n_query_channels=128,
        patch_size=16,
        dim_out=256,
        embedding_dim=128,
        num_heads=4,
        norm="linear",
    ):
        super(mViT, self).__init__()
        self.norm = norm
        self.n_query_channels = n_query_channels
        self.patch_transformer = PatchTransformerEncoder(
            in_channels, patch_size, embedding_dim, num_heads
        )
        self.dot_product_layer = PixelWiseDotProduct()

        self.conv3x3 = nn.Conv2d(
            in_channels, embedding_dim, kernel_size=3, stride=1, padding=1
        )
        self.regressor = nn.Sequential(
            nn.Linear(embedding_dim, 256),
            nn.LeakyReLU(),
            nn.Linear(256, 256),
            nn.LeakyReLU(),
            nn.Linear(256, dim_out),
        )

    def forward(self, x):
        # n, c, h, w = x.size()
        tgt = self.patch_transformer(x.clone())  # .shape = S, N, E

        x = self.conv3x3(x)

        regression_head, queries = tgt[0, ...], tgt[1 : self.n_query_channels + 1, ...]

        # Change from S, N, E to N, S, E
        queries = queries.permute(1, 0, 2)
        range_attention_maps = self.dot_product_layer(
            x, queries
        )  # .shape = n, n_query_channels, h, w

        y = self.regressor(regression_head)  # .shape = N, dim_out
        if self.norm == "linear":
            y = torch.relu(y)
            eps = 0.1
            y = y + eps
        elif self.norm == "softmax":
            return torch.softmax(y, dim=1), range_attention_maps
        else:
            y = torch.sigmoid(y)
        y = y / y.sum(dim=1, keepdim=True)
        return y, range_attention_maps



from pathlib import Path
import torch
from lantern import module_device
import lit

from perceptor.transforms import resize
from perceptor.losses.interface import LossInterface


class LiT(LossInterface):
    def __init__(
        self,
        name="LiT-L16L",
        cache_dir=Path("models"),
    ):
        """
        LiT image-text similarity

        Args:
            model_name (str): Name of the model to load. "LiT-B16B_2" or "LiT-L16L"
            cache_dir (str): Path to the directory where the model is cached.
        """
        super().__init__()
        self.model = lit.LiT(name, cache_dir)
        self.encodings = None
        self.weights = None

    @property
    def device(self):
        return module_device(self)

    def add_texts_(self, texts, weights=None):
        return self.add_encodings_(self.model.encode_texts(texts), weights)

    def add_images_(self, images, weights=None):
        if images.shape[-2:] != self.model.image_size:
            images = resize(images, out_shape=self.model.image_size)
        return self.add_encodings_(self.model.encode_images(images), weights)

    def add_encodings_(
        self,
        encodings,
        weights=None,
    ):
        if isinstance(weights, list) or isinstance(weights, tuple):
            weights = torch.tensor(weights)
        elif weights is None:
            weights = torch.ones_like(encodings[:, 0])

        if self.encodings is None:
            self.encodings = torch.nn.Parameter(
                encodings.to(self.device), requires_grad=False
            )
            self.weights = torch.nn.Parameter(
                weights.to(self.device),
                requires_grad=False,
            )
        else:
            self.encodings = torch.nn.Parameter(
                torch.cat([self.encodings, encodings.to(self.device)]),
                requires_grad=False,
            )
            self.weights = torch.nn.Parameter(
                torch.cat([self.weights, weights.to(self.device)]),
                requires_grad=False,
            )
        return self

    def forward(self, images):
        if images.shape[-2:] != self.model.image_size:
            images = resize(images, out_shape=self.model.image_size)
        image_encodings = self.model.encode_images(images)
        spherical_distance = (
            (image_encodings[:, None] - self.encodings[None, :])
            .norm(dim=2)
            .div(2)
            .arcsin()
            .square()
            .mul(2)
        )
        return (spherical_distance * self.weights).mean()


def test_lit():
    torch.set_grad_enabled(False)
    loss = LiT().add_texts_(["hello", "world"]).add_images_(torch.randn(1, 3, 256, 256))

    image = torch.randn(1, 3, 256, 256).requires_grad_()
    with torch.enable_grad():
        loss(image).backward()

    assert image.grad is not None



import torch
from torch import nn
from torch.nn import functional as F
from basicsr.utils.download_util import load_file_from_url

from .interface import LossInterface
from perceptor import models


class AestheticVisualAssessment(LossInterface):
    def __init__(self, aesthetic_target=10, mode="expected"):
        """
        Aesthetic visual assessment loss based on a classifier that predicts the aesthetic rating of an image.

        Args:
            aesthetic_target (int): Target asthetic rating of the image (1-10).
            mode (str): How to compute the loss. Available modes are:
                - "logit": compute the loss based on the logits of the classifier.
                - "expected": compute the loss based on the expected value of the classifier.
                - "probability": compute the loss based on the probability of the classifier.
        """
        super().__init__()
        self.aesthetic_target = aesthetic_target
        self.mode = mode

        self.model = models.CLIP("ViT-B-16")

        checkpoint_path = load_file_from_url(
            "http://batbot.tv/ai/models/v-diffusion/ava_vit_b_16_full.pth",
            "models",
        )
        self.aesthetic_head = nn.Linear(512, 10)
        self.aesthetic_head.load_state_dict(
            torch.load(checkpoint_path, map_location="cpu")
        )
        self.aesthetic_head.eval()
        self.aesthetic_head.requires_grad_(False)

    def forward(self, images):
        log_probs = self.aesthetic_head(self.model.encode_images(images))
        if self.mode == "logit":
            return -log_probs[..., self.aesthetic_target - 1].mean().mul(0.01)
        elif self.mode == "expected":
            expected_target = F.softmax(log_probs, dim=-1) * torch.arange(10).add(1).to(
                images.device
            )
            return (expected_target - self.aesthetic_target).square().mean().mul(0.01)
        elif self.mode == "probability":
            return -F.softmax(log_probs, dim=-1)[..., self.aesthetic_target - 1].mean()
        else:
            raise ValueError(f"Unknown mode: {self.mode}")


def test_aesthetic_visual_assessment():
    loss = AestheticVisualAssessment().cuda()
    loss(torch.randn(1, 3, 256, 256).cuda())



"""
Highly modified version of https://github.com/shariqfarooq123/AdaBins/blob/main/infer.py
"""
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms
import torchvision.transforms.functional as TF

from . import model_io
from .models import UnetAdaptiveBins


class InferenceHelper(nn.Module):
    def __init__(self, dataset, pretrained_path):
        super().__init__()
        self.normalize = transforms.Normalize(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )
        if dataset == "nyu":
            self.min_depth = 1e-3
            self.max_depth = 10
            self.saving_factor = 1000  # used to save in 16 bit
            model = UnetAdaptiveBins.build(
                n_bins=256, min_val=self.min_depth, max_val=self.max_depth
            )
        elif dataset == "kitti":
            self.min_depth = 1e-3
            self.max_depth = 80
            self.saving_factor = 256
            model = UnetAdaptiveBins.build(
                n_bins=256, min_val=self.min_depth, max_val=self.max_depth
            )
        else:
            raise ValueError(
                f"dataset can be either 'nyu' or 'kitti' but got {dataset}"
            )

        model, _ = model_io.load_checkpoint(pretrained_path, model)
        model.eval()
        model.requires_grad_(False)
        self.model = model

    def predict(self, image):
        """
        Args:
            image: torch.Tensor of shape (1, 3, H, W) between 0 and 1
        """
        bins, pred = self.model(self.normalize(image))
        pred = pred.clamp(self.min_depth, self.max_depth)

        # Flip
        pred_lr = TF.hflip(
            self.model(TF.hflip(image))[-1].clamp(self.min_depth, self.max_depth)
        )

        # Take average of original and mirror
        final = (pred + pred_lr) / 2
        final = F.interpolate(
            final,
            image.shape[-2:],
            mode="bilinear",
            align_corners=True,
        ).clamp(self.min_depth, self.max_depth)

        # final[np.isinf(final)] = self.max_depth
        # final[np.isnan(final)] = self.min_depth

        # centers = 0.5 * (bins[:, 1:] + bins[:, :-1])
        # centers = centers.cpu().squeeze().numpy()
        # centers = centers[centers > self.min_depth]
        # centers = centers[centers < self.max_depth]

        # return centers, final
        return final



from torch import nn

from perceptor.drawers.interface import DrawingInterface
from .get_vae import get_vae


class BruteRuDalle(DrawingInterface):
    def __init__(self, init_images):
        super().__init__()
        self.model = (
            get_vae(dwt=False, cache_dir="models").model.eval().requires_grad_(False)
        )
        self.latent = nn.Parameter(self.encode(init_images), requires_grad=True)

    def synthesize(self, _=None):
        return self.decode(self.latent)

    def encode(self, image):
        quant, embedding_loss, info = self.model.encode(image.mul(2).sub(1))
        return quant

    def decode(self, latent):
        return self.model.decode(latent).add(1).div(2)



from omegaconf import OmegaConf
import torch
import einops
import torch
from basicsr.utils.download_util import load_file_from_url

from perceptor.transforms.resize import resize
from .ldm.util import instantiate_from_config
from perceptor.utils import cache
from . import diffusion_space


@cache
class SuperResolution(torch.nn.Module):
    def __init__(self, eta=1.0, convolutional=False, kernel_size=128, stride=64):
        super().__init__()
        self.eta = eta

        url_conf = "https://heibox.uni-heidelberg.de/f/31a76b13ea27482981b4/?dl=1"
        url_ckpt = "https://heibox.uni-heidelberg.de/f/578df07c8fc04ffbadf3/?dl=1"

        config_path = load_file_from_url(
            url_conf, "models", file_name="sharpen-colab.yaml"
        )
        checkpoint_path = load_file_from_url(
            url_ckpt, "models", file_name="sharpen-colab.ckpt"
        )

        config = OmegaConf.load(config_path)
        self.model = instantiate_from_config(config.model)
        self.model.load_state_dict(
            torch.load(checkpoint_path, map_location="cpu")["state_dict"], strict=False
        )
        self.model.requires_grad_(False)
        self.model.cuda()
        self.model.eval()

        self.up_f = 4
        if convolutional:
            vqf = 4
            self.model.split_input_params = {
                "ks": (kernel_size, kernel_size),
                "stride": (stride, stride),
                "vqf": vqf,
                "patch_distributed_vq": True,
                "tie_braker": False,
                "clip_max_weight": 0.5,
                "clip_min_weight": 0.01,
                "clip_max_tie_weight": 0.5,
                "clip_min_tie_weight": 0.01,
            }
        else:
            if hasattr(self.model, "split_input_params"):
                delattr(self.model, "split_input_params")

    @property
    def device(self):
        return next(iter(self.parameters())).device

    def forward(self, latents, conditioning, index):
        return self.denoise(latents, conditioning, index)

    def velocity(self, latents, conditioning, index):
        raise NotImplementedError()

    def upsample(self, images):
        return resize(images, out_shape=[s * self.up_f for s in images.shape[-2:]])

    def latents(self, images):
        """Encode images (0-1) to latent space"""
        example = dict(
            LR_image=diffusion_space.encode(
                einops.rearrange(
                    resize(
                        images, out_shape=[s // self.up_f for s in images.shape[-2:]]
                    ),
                    "1 c h w -> 1 h w c",
                )
            ),
            image=einops.rearrange(images, "1 c h w -> 1 h w c"),
        )

        latents, conditioning = self.model.get_input(
            example,
            self.model.first_stage_key,
            return_first_stage_outputs=False,
            force_c_encode=not (
                hasattr(self.model, "split_input_params")
                and self.model.cond_stage_key == "coordinates_bbox"
            ),
            return_original_cond=False,
        )
        return latents

    def conditioning(self, images):
        latents, conditioning = self.latents(images)
        return conditioning

    def diffuse(self, latents, index, noise=None):
        if noise is None:
            noise = torch.randn_like(latents)

        return self.model.q_sample(x_start=latents, t=self.ts(index), noise=noise)

    def denoise(self, latents, conditioning, index, eps=None):
        """Predict denoised latents"""
        if index >= 1000:
            raise ValueError("index must be less than 1000")

        if eps is None:
            eps = self.eps(latents, index, conditioning)

        return (
            latents - self.sqrt_one_minus_alphas_cumprod(index) * eps
        ) / self.alphas_cumprod(index).sqrt()

    def images(self, latents):
        """Decode from latent space to image"""
        return diffusion_space.decode(self.model.decode_first_stage(latents))

    def ts(self, index):
        return torch.tensor([index], device=self.device, dtype=torch.long)

    def alphas_cumprod(self, index):
        return self.model.alphas_cumprod[index].to(self.device)[None, None, None, None]

    def sqrt_one_minus_alphas_cumprod(self, index):
        return self.model.sqrt_one_minus_alphas_cumprod[index].to(self.device)[
            None, None, None, None
        ]

    def step(
        self, from_latents, predicted_denoised_latents, from_index, to_index, noise=None
    ):
        if to_index > from_index:
            raise ValueError("to_index must be smaller than from_index")
        if noise is None:
            noise = torch.randn_like(predicted_denoised_latents)

        # if quantize_denoised:
        #     predicted_denoised_latents, _, *_ = self.model.first_stage_model.quantize(predicted_denoised_latents)

        from_alphas_cumprod = self.alphas_cumprod(from_index)
        to_alphas_cumprod = self.alphas_cumprod(to_index)
        from_sqrt_one_minus_alphas_cumprod = self.sqrt_one_minus_alphas_cumprod(
            from_index
        )

        to_sigmas = self.eta * torch.sqrt(
            (1 - to_alphas_cumprod)
            / (1 - from_alphas_cumprod)
            * (1 - from_alphas_cumprod / to_alphas_cumprod)
        )

        eps = (
            from_latents - predicted_denoised_latents * from_alphas_cumprod.sqrt()
        ) / from_sqrt_one_minus_alphas_cumprod

        dir_xt = (1.0 - to_alphas_cumprod - to_sigmas**2).sqrt() * eps

        to_z = (
            to_alphas_cumprod.sqrt() * predicted_denoised_latents
            + dir_xt
            + to_sigmas * noise
        )
        return to_z

    def eps(self, latents, index, conditioning):
        if index >= 1000:
            raise ValueError("index must be less than 1000")

        return self.model.apply_model(latents, self.ts(index), conditioning)



def get_features(image, model, layers=None):

    if layers is None:
        layers = {
            "0": "conv1_1",
            "5": "conv2_1",
            "10": "conv3_1",
            "19": "conv4_1",
            "21": "conv4_2",
            "28": "conv5_1",
            "31": "conv5_2",
        }
    features = {}
    x = image
    for name, layer in model._modules.items():
        x = layer(x)
        if name in layers:
            features[layers[name]] = x

    return features



from typing import Optional
from tqdm import tqdm
from scipy import integrate
import torch
import torch.nn as nn
import torch.nn.functional as F
from lantern import module_device, Tensor
from basicsr.utils.download_util import load_file_from_url

from . import settings
from . import standardize
from . import base
from . import diffusion
from .prediction import PredictionBatch

ALL_CHECKPOINT_URL = "https://s3.eu-central-1.wasabisys.com/nextml-model-data/monster-diffusion/6b70ff1e6c7f4c00ad8cb59879f7d88d.pt"
TINY_HERO_CHECKPOINT_URL = "https://s3.eu-central-1.wasabisys.com/nextml-model-data/monster-diffusion/f47af8975b744d4bae2b905bac223003.pt"


class MonsterDiffusion(nn.Module):
    def __init__(self, name="all"):
        super().__init__()
        self.network = base.Model(
            mapping_cond_dim=9,
        )
        if name == "all":
            checkpoint_path = load_file_from_url(ALL_CHECKPOINT_URL, "models")
        elif name == "tiny-hero":
            checkpoint_path = load_file_from_url(TINY_HERO_CHECKPOINT_URL, "models")
        else:
            raise ValueError(f"Unknown model name {name}")
        self.load_state_dict(torch.load(checkpoint_path))
        self.eval().requires_grad_(False)

    @property
    def device(self):
        return module_device(self)

    @staticmethod
    def training_ts(size):
        random_ts = (diffusion.P_mean + torch.randn(size) * diffusion.P_std).exp()
        return random_ts

    def _schedule_ts(self, n_steps):
        ramp = torch.linspace(0, 1, n_steps).to(self.device)
        min_inv_rho = diffusion.sigma_min ** (1 / diffusion.rho)
        max_inv_rho = diffusion.sigma_max ** (1 / diffusion.rho)
        return (max_inv_rho + ramp * (min_inv_rho - max_inv_rho)) ** diffusion.rho

    def schedule_ts(self, n_steps):
        schedule_ts = self._schedule_ts(n_steps)
        return zip(schedule_ts[:-1], schedule_ts[1:])

    @staticmethod
    def sigmas(ts):
        return PredictionBatch.sigmas(ts)

    @staticmethod
    def alphas(ts):
        return PredictionBatch.alphas(ts)

    @staticmethod
    def random_noise(size):
        return standardize.decode(
            torch.randn(size, *settings.INPUT_SHAPE) * diffusion.sigma_max
        )

    @staticmethod
    def diffuse(
        images: Tensor.dims("NCHW").shape(-1, *settings.INPUT_SHAPE).float(),
        ts,
        noise=None,
    ):
        x0 = standardize.encode(images)
        if isinstance(ts, float) or ts.ndim == 0:
            ts = torch.full((x0.shape[0],), ts).to(x0.device)

        if noise is None:
            noise = torch.randn_like(x0).to(x0.device)

        assert x0.shape == noise.shape

        return standardize.decode(x0 + noise * MonsterDiffusion.sigmas(ts))

    def c_skip(self, ts):
        return diffusion.sigma_data**2 / (
            diffusion.sigma_data**2 + self.sigmas(ts) ** 2
        )

    def c_out(self, ts):
        return (
            self.sigmas(ts)
            * diffusion.sigma_data
            / torch.sqrt(diffusion.sigma_data**2 + self.sigmas(ts) ** 2)
        )

    def c_in(self, ts):
        return 1 / torch.sqrt(diffusion.sigma_data**2 + self.sigmas(ts) ** 2)

    def c_noise(self, ts):
        return 1 / 4 * self.sigmas(ts).log().view(-1)

    def denoised_(
        self,
        diffused_images: Tensor.dims("NCHW").shape(-1, *settings.INPUT_SHAPE).float(),
        ts: Tensor.dims("N"),
        nonleaky_augmentations: Optional[Tensor.dims("NK")] = None,
    ) -> Tensor.dims("NCHW").shape(-1, *settings.INPUT_SHAPE):
        """
        Parameterization from https://arxiv.org/pdf/2206.00364.pdf
        """
        diffused_xs = standardize.encode(diffused_images.to(self.device))
        ts = ts.to(self.device)
        if nonleaky_augmentations is None:
            nonleaky_augmentations = torch.zeros(
                len(diffused_images), settings.N_AUGMENTATIONS
            )
        nonleaky_augmentations = nonleaky_augmentations.to(self.device)

        output = self.network(
            self.c_in(ts) * diffused_xs,
            self.c_noise(self.sigmas(ts).flatten()),
            mapping_cond=nonleaky_augmentations,
        )
        return self.c_skip(ts) * diffused_xs + self.c_out(ts) * output

    def forward(
        self,
        diffused_images: Tensor.dims("NCHW"),
        ts: Tensor.dims("N"),
        nonleaky_augmentations: Optional[Tensor.dims("NK")] = None,
    ):
        denoised_xs = self.denoised_(
            diffused_images,
            ts,
            nonleaky_augmentations,
        )
        return PredictionBatch(
            denoised_xs=denoised_xs,
            diffused_images=diffused_images,
            ts=torch.as_tensor(ts).flatten().to(self.device),
        )

    def predictions_(
        self,
        diffused_images: Tensor.dims("NCHW"),
        ts: Tensor.dims("N"),
        nonleaky_augmentations: Optional[Tensor.dims("NK")] = None,
    ):
        return self.forward(
            diffused_images,
            ts,
            nonleaky_augmentations,
        )

    def predictions(
        self,
        diffused_images: Tensor.dims("NCHW"),
        ts: Tensor.dims("N"),
        nonleaky_augmentations: Optional[Tensor.dims("NK")] = None,
    ):
        if self.training:
            raise Exception(
                "Cannot run predictions method while in training mode. Use predictions_"
            )
        return self.predictions_(
            diffused_images,
            ts,
            nonleaky_augmentations,
        )

    @staticmethod
    def gamma(ts, n_steps):
        return torch.where(
            (ts >= diffusion.S_tmin) & (ts <= diffusion.S_tmax),
            torch.minimum(
                torch.tensor(diffusion.S_churn / n_steps),
                torch.tensor(2).sqrt() - 1,
            ).to(ts),
            torch.zeros_like(ts),
        )

    @staticmethod
    def reversed_ts(ts, n_steps):
        return ts + MonsterDiffusion.gamma(ts, n_steps) * ts

    def inject_noise(self, diffused_images, ts, reversed_ts):
        diffused_xs = standardize.encode(diffused_images).to(self.device)

        reversed_diffused_xs = (
            diffused_xs
            + (self.sigmas(reversed_ts).square() - self.sigmas(ts).square()).sqrt()
            * torch.randn_like(diffused_xs)
            * diffusion.S_noise
        )
        return standardize.decode(reversed_diffused_xs)

    def sample(
        self,
        size,
        n_evaluations=100,
        progress=False,
        diffused_images=None,
    ):
        return self.elucidated_sample(
            size,
            n_evaluations,
            progress,
            diffused_images,
        )

    def elucidated_sample(
        self,
        size,
        n_evaluations=100,
        progress=False,
        diffused_images=None,
    ):
        """
        Elucidated stochastic sampling from https://arxiv.org/pdf/2206.00364.pdf
        """
        if self.training:
            raise Exception("Cannot run sample method while in training mode.")
        if diffused_images is None:
            diffused_images = self.random_noise(size).to(self.device)
        nonleaky_augmentations = torch.zeros(
            (size, settings.N_AUGMENTATIONS), dtype=torch.float32, device=self.device
        )

        n_steps = n_evaluations // 2
        i = 0
        progress = tqdm(total=n_steps, disable=not progress, leave=False)
        for from_ts, to_ts in self.schedule_ts(n_steps):
            reversed_ts = self.reversed_ts(from_ts, n_steps).clamp(
                max=diffusion.sigma_max
            )
            reversed_diffused_images = self.inject_noise(
                diffused_images, from_ts, reversed_ts
            )
            i += 1

            predictions = self.predictions(
                reversed_diffused_images,
                reversed_ts,
                nonleaky_augmentations,
            )
            reversed_eps = predictions.eps
            diffused_images = predictions.step(to_ts)

            predictions = self.predictions(
                diffused_images,
                to_ts,
                nonleaky_augmentations,
            )
            diffused_images = predictions.correction(
                reversed_diffused_images, reversed_ts, reversed_eps
            )
            progress.update()
            yield predictions.denoised_images.clamp(0, 1)

        reversed_ts = self.reversed_ts(to_ts, n_steps)
        diffused_images = self.inject_noise(diffused_images, to_ts, reversed_ts)

        predictions = self.predictions(
            diffused_images,
            reversed_ts,
            nonleaky_augmentations,
        )
        progress.close()
        yield predictions.denoised_images.clamp(0, 1)

    @staticmethod
    def linear_multistep_coeff(order, sigmas, from_index, to_index):
        if order - 1 > from_index:
            raise ValueError(f"Order {order} too high for step {from_index}")

        def fn(tau):
            prod = 1.0
            for k in range(order):
                if to_index == k:
                    continue
                prod *= (tau - sigmas[from_index - k]) / (
                    sigmas[from_index - to_index] - sigmas[from_index - k]
                )
            return prod

        return integrate.quad(
            fn, sigmas[from_index], sigmas[from_index + 1], epsrel=1e-4
        )[0]

    def linear_multistep_sample(
        self,
        size,
        n_evaluations=100,
        progress=False,
        diffused_images=None,
        order=4,
    ):
        """
        Katherine Crowson's linear multistep method from https://github.com/crowsonkb/k-diffusion/blob/4fdb34081f7a09f16c33d3344a042e5bea8e69ee/k_diffusion/sampling.py
        """
        if self.training:
            raise Exception("Cannot run sample method while in training mode.")
        if diffused_images is None:
            diffused_images = self.random_noise(size)
        nonleaky_augmentations = torch.zeros(
            (size, settings.N_AUGMENTATIONS), dtype=torch.float32, device=self.device
        )
        diffused_images = diffused_images.to(self.device)

        n_steps = n_evaluations
        schedule_ts = self._schedule_ts(n_steps)

        epses = list()
        progress = tqdm(total=n_steps, disable=not progress, leave=False)
        for from_index, (from_ts, to_ts) in enumerate(self.schedule_ts(n_steps)):

            predictions = self.predictions(
                diffused_images,
                from_ts,
                nonleaky_augmentations,
            )
            epses.append(predictions.eps)
            if len(epses) > order:
                epses.pop(0)

            current_order = len(epses)
            coeffs = [
                self.linear_multistep_coeff(
                    current_order,
                    self.sigmas(schedule_ts).cpu().flatten(),
                    from_index,
                    to_index,
                )
                for to_index in range(current_order)
            ]

            diffused_xs = standardize.encode(diffused_images)
            diffused_xs = diffused_xs + sum(
                coeff * eps for coeff, eps in zip(coeffs, reversed(epses))
            )
            diffused_images = standardize.decode(diffused_xs)

            progress.update()
            yield predictions.denoised_images.clamp(0, 1)

        predictions = self.predictions(
            diffused_images,
            to_ts,
            nonleaky_augmentations,
        )
        progress.close()
        yield predictions.denoised_images.clamp(0, 1)


def test_monster_diffusion():
    from perceptor import utils

    model = MonsterDiffusion().cuda()
    for images in model.sample(size=1, n_evaluations=50):
        pass
    utils.pil_image(images).save("tests/monster_diffusion.png")


def test_monster_diffusion_lms():
    from perceptor import utils

    model = MonsterDiffusion().cuda()
    for images in model.linear_multistep_sample(size=1, n_evaluations=50):
        pass
    utils.pil_image(images).save("tests/monster_diffusion_lms.png")



from typing import List, Union, Optional
import numpy as np
import torch
import torch.nn.functional as F


resume_specs = {
    # For StyleGAN2/ADA models; --cfg=stylegan2
    "stylegan2": {
        # Official NVIDIA models
        "ffhq256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhq-256x256.pkl",
        "ffhqu256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhqu-256x256.pkl",
        "ffhq512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhq-512x512.pkl",
        "ffhq1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhq-1024x1024.pkl",
        "ffhqu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhqu-1024x1024.pkl",
        "celebahq256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-celebahq-256x256.pkl",
        "lsundog256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-lsundog-256x256.pkl",
        "afhqcat512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-afhqcat-512x512.pkl",
        "afhqdog512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-afhqdog-512x512.pkl",
        "afhqwild512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-afhqwild-512x512.pkl",
        "afhq512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-afhqv2-512x512.pkl",
        "brecahad512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-brecahad-512x512.pkl",
        "cifar10": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-cifar10-32x32.pkl",
        "metfaces1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-metfaces-1024x1024.pkl",
        "metfacesu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-metfacesu-1024x1024.pkl",
        # Community models; TODO: add the interesting ones found in https://github.com/justinpinkney/awesome-pretrained-stylegan2
        "minecraft1024": "https://github.com/jeffheaton/pretrained-gan-minecraft/releases/download/v1/minecraft-gan-2020-12-22.pkl",  # Thanks to @jeffheaton
        # Deceive-D/APA models (ignoring the faces models): https://github.com/EndlessSora/DeceiveD
        "afhqcat256": "https://drive.google.com/u/0/uc?export=download&confirm=zFoN&id=1P9ouHIK-W8JTb6bvecfBe4c_3w6gmMJK",
        "anime256": "https://drive.google.com/u/0/uc?export=download&confirm=6Uie&id=1EWOdieqELYmd2xRxUR4gnx7G10YI5dyP",
        "cub256": "https://drive.google.com/u/0/uc?export=download&confirm=KwZS&id=1J0qactT55ofAvzddDE_xnJEY8s3vbo1_",
    },
    # For StyleGAN3 config-r models (--cfg=stylegan3-r)
    "stylegan3-r": {
        # Official NVIDIA models
        "afhq512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-afhqv2-512x512.pkl",
        "ffhq1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-ffhq-1024x1024.pkl",
        "ffhqu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-ffhqu-1024x1024.pkl",
        "ffhqu256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-ffhqu-256x256.pkl",
        "metfaces1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-metfaces-1024x1024.pkl",
        "metfacesu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-metfacesu-1024x1024.pkl",
    },
    # For StyleGAN3 config-t models (--cfg=stylegan3-t)
    "stylegan3-t": {
        # Official NVIDIA models
        "afhq512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-afhqv2-512x512.pkl",
        "ffhq1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-ffhq-1024x1024.pkl",
        "ffhqu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-ffhqu-1024x1024.pkl",
        "ffhqu256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-ffhqu-256x256.pkl",
        "metfaces1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-metfaces-1024x1024.pkl",
        "metfacesu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-metfacesu-1024x1024.pkl",
        # Community models, found in: https://github.com/justinpinkney/awesome-pretrained-stylegan3 by @justinpinkney
        "landscapes256": "https://drive.google.com/u/0/uc?export=download&confirm=eJHe&id=14UGDDOusZ9TMb-pOrF0PAjMGVWLSAii1",  # Thanks to @justinpinkney
        "wikiart1024": "https://drive.google.com/u/0/uc?export=download&confirm=2tz5&id=18MOpwTMJsl_Z17q-wQVnaRLCUFZYSNkj",  # Thanks to @justinpinkney
    },
}


def z_to_img(
    G,
    latents: torch.Tensor,
    label: torch.Tensor,
    truncation_psi: float,
    noise_mode: str = "const",
) -> np.ndarray:
    """
    Get an image/np.ndarray from a latent Z using G, the label, truncation_psi, and noise_mode. The shape
    of the output image/np.ndarray will be [len(dlatents), G.img_resolution, G.img_resolution, G.img_channels]
    """
    assert isinstance(
        latents, torch.Tensor
    ), f'latents should be a torch.Tensor!: "{type(latents)}"'
    if len(latents.shape) == 1:
        latents = latents.unsqueeze(0)  # An individual latent => [1, G.z_dim]
    img = G(z=latents, c=label, truncation_psi=truncation_psi, noise_mode=noise_mode)
    return img


def w_to_img(
    G,
    dlatents: Union[List[torch.Tensor], torch.Tensor],
    noise_mode: str = "const",
) -> np.ndarray:
    """
    Get an image/np.ndarray from a dlatent W using G and the selected noise_mode. The final shape of the
    returned image will be [len(dlatents), G.img_resolution, G.img_resolution, G.img_channels].
    """
    assert isinstance(
        dlatents, torch.Tensor
    ), f'dlatents should be a torch.Tensor!: "{type(dlatents)}"'
    if len(dlatents.shape) == 2:
        dlatents = dlatents.unsqueeze(
            0
        )  # An individual dlatent => [1, G.mapping.num_ws, G.mapping.w_dim]
    synth_image = G.synthesis(dlatents, noise_mode=noise_mode)
    return synth_image


def get_w_from_seed(
    G,
    batch_sz: int,
    device: torch.device,
    truncation_psi: float,
    seed: Optional[int],
    # centroids_path: Optional[str],
    class_idx: Optional[int],
) -> torch.Tensor:
    """Get the dlatent from a list of random seeds, using the truncation trick (this could be optional)"""

    if G.c_dim != 0:
        # sample random labels if no class idx is given
        if class_idx is None:
            class_indices = np.random.RandomState(seed).randint(
                low=0, high=G.c_dim, size=(batch_sz)
            )
            class_indices = torch.from_numpy(class_indices).to(device).long()
            w_avg = G.mapping.w_avg.index_select(0, class_indices)
        else:
            w_avg = G.mapping.w_avg[class_idx].unsqueeze(0).repeat(batch_sz, 1)
            class_indices = torch.full((batch_sz,), class_idx).to(device)

        labels = F.one_hot(class_indices, G.c_dim)

    else:
        w_avg = G.mapping.w_avg.unsqueeze(0)
        labels = None
        if class_idx is not None:
            print("Warning: --class is ignored when running an unconditional network")

    z = np.random.RandomState(seed).randn(batch_sz, G.z_dim)
    z = torch.from_numpy(z).to(device)
    w = G.mapping(z, labels)

    w_avg = w_avg.unsqueeze(1).repeat(1, G.mapping.num_ws, 1)
    w = w_avg + (w - w_avg) * truncation_psi

    return w



import torch


def normalize(image):
    mean = torch.tensor([0.485, 0.456, 0.406]).to(image.device)
    std = torch.tensor([0.229, 0.224, 0.225]).to(image.device)
    mean = mean.view(1, -1, 1, 1)
    std = std.view(1, -1, 1, 1)

    image = (image - mean) / std
    return image



from typing import Optional

from .open_clip import OpenCLIP


def CLIP(architecture: str, precision: Optional[str] = None, jit=True):
    """
    Args:
        architecture (str): name of the clip model. Available models are:
            - RN50 [-quickgelu]
            - RN101 [-quickgelu]
            - RN50x4
            - RN50x16
            - RN50x64
            - ViT-B-32 [-quickgelu]
            - ViT-B-16
            - ViT-L-14
            - ViT-L-14-336px
        precision (str): precision of the model. Options are "fp32" and "fp16"
    """
    if "-quickgelu" not in architecture and architecture in [
        "RN50",
        "RN101",
        "ViT-B-32",
    ]:
        architecture = architecture + "-quickgelu"
    return OpenCLIP(architecture, "openai", precision, jit=jit)


def test_clip():
    import torch

    model = CLIP("ViT-B-32")

    image = torch.randn((1, 3, 256, 256)).requires_grad_()
    with torch.enable_grad():
        model.encode_images(image).mean().backward()

    assert image.grad is not None



import torch
import lantern

from . import diffusion_space, utils
from perceptor.transforms.clamp_with_grad import clamp_with_grad


class Predictions(lantern.FunctionalBase):
    from_diffused_images: lantern.Tensor.dims("NCHW")
    from_ts: lantern.Tensor.dims("N")
    velocities: lantern.Tensor.dims("NCHW")

    @property
    def device(self):
        return self.velocities.device

    def alphas(self, ts):
        if isinstance(ts, float):
            ts = torch.tensor(ts)
        if ts.ndim == 0:
            ts = ts[None]
        if ts.ndim != 1:
            raise ValueError("ts must be a scalar or a 1D tensor")
        alphas, _ = utils.t_to_alpha_sigma(ts)
        return alphas[:, None, None, None].to(self.device)

    def sigmas(self, ts):
        if isinstance(ts, float):
            ts = torch.tensor(ts)
        if ts.ndim == 0:
            ts = ts[None]
        if ts.ndim != 1:
            raise ValueError("ts must be a scalar or a 1D tensor")
        _, sigmas = utils.t_to_alpha_sigma(ts)
        return sigmas[:, None, None, None].to(self.device)

    @property
    def from_alphas(self):
        return self.alphas(self.from_ts)

    @property
    def from_sigmas(self):
        return self.sigmas(self.from_ts)

    @property
    def from_diffused_xs(self):
        return diffusion_space.encode(self.from_diffused_images)

    @property
    def denoised_xs(self):
        return (
            self.from_diffused_xs * self.from_alphas
            - self.velocities * self.from_sigmas
        )

    @property
    def predicted_noise(self):
        return (
            self.from_diffused_xs * self.from_sigmas
            + self.velocities * self.from_alphas
        )

    @property
    def denoised_images(self):
        return diffusion_space.decode(self.denoised_xs)

    def step(self, to_ts, eta=0.0):
        """
        Reduce noise level to `to_ts`

        Args:
            to_ts: Union[Tensor, Tensor.shape("N"), float]
            eta: float

        Returns:
            diffused_images: torch.Tensor.shape("NCHW")
        """
        to_alphas, to_sigmas = self.alphas(to_ts), self.sigmas(to_ts)

        if eta > 0.0:
            # If eta > 0, adjust the scaling factor for the predicted noise
            # downward according to the amount of additional noise to add
            ddim_sigma = (
                eta
                * (to_sigmas**2 / self.from_sigmas**2).sqrt()
                * (1 - self.from_alphas**2 / to_alphas**2).sqrt()
            )
            adjusted_sigma = (to_sigmas**2 - ddim_sigma**2).sqrt()

            # Recombine the predicted noise and predicted denoised image in the
            # correct proportions for the next step
            to_diffused_xs = (
                self.denoised_xs * to_alphas + self.predicted_noise * adjusted_sigma
            )

            # Add the correct amount of fresh noise
            noise = torch.randn_like(to_diffused_xs)
            to_diffused_xs += noise * ddim_sigma
        else:
            to_diffused_xs = (
                self.denoised_xs * to_alphas + self.predicted_noise * to_sigmas
            )

        return diffusion_space.decode(to_diffused_xs)

    def correction(self, previous: "Predictions"):
        return previous.forced_denoised(
            (self.denoised_images + previous.denoised_images) / 2
        )

    def reverse_step(self, to_ts):
        if (torch.as_tensor(self.from_ts) > torch.as_tensor(to_ts)).any():
            raise ValueError("from_ts must be less than to_ts")

        to_alphas, to_sigmas = self.alphas(to_ts), self.sigmas(to_ts)
        return self.denoised_xs * to_alphas + self.predicted_noise * to_sigmas

    def resample_noise(self, resample_ts):
        if (torch.as_tensor(self.from_ts) < torch.as_tensor(resample_ts)).any():
            raise ValueError("from_ts must be greater than resample_ts")
        resampled_noise_sigma = (
            self.sigmas(resample_ts) * self.predicted_noise
            + (
                self.from_sigmas**2 - self.sigmas(resample_ts) ** 2
            ).sqrt() * torch.randn_like(self.predicted_noise)
        )  # fmt: skip
        return resampled_noise_sigma / self.from_sigmas

    def resample(self, resample_ts):
        """
        Harmonizing resampling from https://github.com/andreas128/RePaint
        """
        return diffusion_space.decode(
            self.denoised_xs * self.from_alphas
            + self.resample_noise(resample_ts) * self.from_sigmas
        )

    def noisy_reverse_step(self, to_ts):
        to_alphas, to_sigmas = self.alphas(to_ts), self.sigmas(to_ts)

        noise_sigma = self.from_sigmas * self.predicted_noise + (
            to_sigmas**2 - self.from_sigmas**2
        ).sqrt() * torch.randn_like(self.predicted_noise)

        return diffusion_space.decode(self.denoised_xs * to_alphas + noise_sigma)

    def guided(self, guiding, guidance_scale=0.5, clamp_value=1e-6) -> "Predictions":
        return self.replace(
            velocities=self.velocities
            + guidance_scale
            * self.from_sigmas
            * guiding.clamp(-clamp_value, clamp_value)
            / clamp_value
        )

    def dynamic_threshold(self, quantile=0.95) -> "Predictions":
        """
        Thresholding heuristic from imagen paper
        """
        dynamic_threshold = torch.quantile(
            self.denoised_xs.flatten(start_dim=1).abs(), quantile, dim=1
        ).clamp(min=1.0)
        denoised_xs = (
            clamp_with_grad(
                self.denoised_xs,
                -dynamic_threshold,
                dynamic_threshold,
            )
            # / dynamic_threshold
            # imagen's dynamic thresholding divides by threshold but this makes the images gray
        )
        return self.forced_denoised(diffusion_space.decode(denoised_xs))

    def static_threshold(self):
        return self.forced_denoised(clamp_with_grad(self.denoised_images, 0, 1))

    def forced_denoised(self, denoised_images) -> "Predictions":
        denoised_xs = diffusion_space.encode(denoised_images)
        if (self.from_sigmas >= 1e-3).all():
            predicted_noise = (
                self.from_diffused_xs - denoised_xs * self.from_alphas
            ) / self.from_sigmas
        else:
            predicted_noise = self.predicted_noise
        return self.replace(
            velocities=self.from_alphas * predicted_noise
            - self.from_sigmas * denoised_xs
        )

    def forced_predicted_noise(self, predicted_noise) -> "Predictions":
        if (self.from_alphas >= 1e-3).all():
            denoised_xs = (
                self.from_diffused_xs - predicted_noise * self.from_sigmas
            ) / self.from_alphas
        else:
            denoised_xs = self.denoised_xs
        return self.replace(
            velocities=self.from_alphas * predicted_noise
            - self.from_sigmas * denoised_xs
        )

    def wasserstein_distance(self):
        sorted_noise = self.predicted_noise.flatten(start_dim=1).sort(dim=1)[0]
        n = sorted_noise.shape[1]
        margin = 0.5 / n
        points = torch.linspace(margin, 1 - margin, sorted_noise.shape[1])
        expected_noise = torch.distributions.Normal(0, 1).icdf(points)
        return (sorted_noise - expected_noise[None].to(sorted_noise)).abs().mean()

    def wasserstein_square_distance(self):
        sorted_noise = self.predicted_noise.flatten(start_dim=1).sort(dim=1)[0]
        n = sorted_noise.shape[1]
        margin = 0.5 / n
        points = torch.linspace(margin, 1 - margin, sorted_noise.shape[1])
        expected_noise = torch.distributions.Normal(0, 1).icdf(points)
        return (sorted_noise - expected_noise[None].to(sorted_noise)).square().mean()



import torch
import torch.nn.functional as F
from torchvision import transforms
from basicsr.utils.download_util import load_file_from_url

from perceptor import utils
from perceptor.transforms.resize import resize
from .blip_itm import blip_itm


checkpoints = {
    "model_base_retrieval_coco": "https://storage.googleapis.com/sfr-vision-language-research/BLIP/models/model_base_retrieval_coco.pth",
    "model_large_retrieval_coco": "https://storage.googleapis.com/sfr-vision-language-research/BLIP/models/model_large_retrieval_coco.pth",
    "model_base_retrieval_flickr": "https://storage.googleapis.com/sfr-vision-language-research/BLIP/models/model_base_retrieval_flickr.pth",
    "model_large_retrieval_flickr": "https://storage.googleapis.com/sfr-vision-language-research/BLIP/models/model_large_retrieval_flickr.pth",
    "model_large": "https://storage.googleapis.com/sfr-vision-language-research/BLIP/models/model_large.pth",
    "model*_base": "https://storage.googleapis.com/sfr-vision-language-research/BLIP/models/model*_base.pth",
    "model_base": "https://storage.googleapis.com/sfr-vision-language-research/BLIP/models/model_base.pth",
    "model_base_capfilt_large": "https://storage.googleapis.com/sfr-vision-language-research/BLIP/models/model_base_capfilt_large.pth",
}

vit_map = {
    "model_base_retrieval_coco": "base",
    "model_large_retrieval_coco": "large",
    "model_base_retrieval_flickr": "base",
    "model_large_retrieval_flickr": "large",
    "model_large": "large",
    "model*_base": "base",
    "model_base": "base",
    "model_base_capfilt_large": "base",
}

image_size = {
    "model_base_retrieval_coco": 384,
    "model_large_retrieval_coco": 384,
    "model_base_retrieval_flickr": 384,
    "model_large_retrieval_flickr": 384,
    "model_large": 384,
    "model*_base": 384,
    "model_base": 224,
    "model_base_capfilt_large": 384,
}


@utils.cache
class BLIP(torch.nn.Module):
    def __init__(self, name="model_base_retrieval_flickr"):
        """
        Args:
            name (str): Name of the model. Available models are:
                - model_base_retrieval_coco
                - model_large_retrieval_coco
                - model_base_retrieval_flickr
                - model_large_retrieval_flickr
                - model_large
                - model*_base
                - model_base
                - model_base_capfilt_large
        """
        super().__init__()
        self.name = name

        checkpoint_path = load_file_from_url(
            checkpoints[self.name],
            "models",
        )

        self.image_size = image_size[name]
        self.model = (
            blip_itm(
                pretrained=checkpoint_path,
                image_size=self.image_size,
                vit=vit_map[self.name],
            )
            .requires_grad_(False)
            .eval()
        )

        self.normalize = transforms.Normalize(
            (0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)
        )

    def encode_texts(self, texts):
        tokenized_texts = self.model.tokenizer(
            texts,
            padding="max_length",
            truncation=True,
            max_length=35,
            return_tensors="pt",
        )
        text_output = self.model.text_encoder(
            tokenized_texts.input_ids,
            attention_mask=tokenized_texts.attention_mask,
            return_dict=True,
            mode="text",
        )
        text_encodings = F.normalize(
            self.model.text_proj(text_output.last_hidden_state[:, 0, :]),
            dim=-1,
        )
        return F.normalize(text_encodings, dim=-1)

    def encode_images(self, images):
        if images.shape[-2:] != (self.image_size, self.image_size):
            images = resize(images, out_shape=(self.image_size, self.image_size))

        image_encodings = self.model.visual_encoder(self.normalize(images))

        image_encodings_itc = F.normalize(
            self.model.vision_proj(image_encodings[:, 0, :]), dim=-1
        )

        return F.normalize(image_encodings_itc, dim=-1)

    def image_text_contrastive_spherical_distance(self, encodings_a, encodings_b):
        return (
            (encodings_a[None, :] - encodings_b[:, None])
            .norm(dim=-1)
            .div(2)
            .arcsin()
            .square()
            .mul(2)
        )

    def forward(self, _):
        raise NotImplementedError



# -*- coding: utf-8 -*-
from os.path import dirname, abspath, join

import torch
from huggingface_hub import hf_hub_url, cached_download
from omegaconf import OmegaConf

from .model import VQGanGumbelVAE


def get_vae(pretrained=True, dwt=False, cache_dir="models"):
    config = OmegaConf.load(
        join(dirname(abspath(__file__)), "vqgan.gumbelf8-sber.config.yml")
    )
    vae = VQGanGumbelVAE(config, dwt=dwt)
    if pretrained:
        repo_id = "shonenkov/rudalle-utils"
        if dwt:
            filename = "vqgan.gumbelf8-sber-dwt.model.ckpt"
        else:
            filename = "vqgan.gumbelf8-sber.model.ckpt"
        cache_dir = join(cache_dir, "vae")
        config_file_url = hf_hub_url(repo_id=repo_id, filename=filename)
        cached_download(config_file_url, cache_dir=cache_dir, force_filename=filename)
        checkpoint = torch.load(join(cache_dir, filename), map_location="cpu")
        if dwt:
            vae.load_state_dict(checkpoint["state_dict"])
        else:
            vae.model.load_state_dict(checkpoint["state_dict"], strict=False)
    return vae



from typing import List
import torch
import torch.nn as nn
import torchvision.transforms.functional as TF
from lantern import Tensor, FunctionalBase
from transformers import OwlViTProcessor
from transformers.tokenization_utils_base import BatchEncoding

from .modeling_owlvit import OwlViTForObjectDetection
from .feaure_extraction_owlvit import OwlViTFeatureExtractor
from perceptor import utils
from perceptor.transforms.resize import resize


class OWLViTEncodings(nn.Module):
    texts: List[List[str]]
    batch_encoding: BatchEncoding

    def __init__(self, texts, batch_encoding):
        super().__init__()
        self.texts = texts
        self.batch_encoding = batch_encoding

    def to(self, device):
        self.batch_encoding.to(device)
        return super().to(device)

    def repeat(self, repetitions):
        return OWLViTEncodings(
            texts=self.texts,
            batch_encoding=BatchEncoding(
                {
                    key: torch.cat([value for _ in range(repetitions)], dim=0)
                    for key, value in self.batch_encoding.items()
                }
            ),
        )


class OWLViTPredictions(FunctionalBase):
    logits: Tensor.dims("NKE")
    boxes: Tensor.dims("NK4")
    scores: Tensor.dims("NK")
    labels: Tensor.dims("NK")
    texts: List[List[str]]


@utils.cache
class OWLViT(torch.nn.Module):
    def __init__(self):
        """
        OWL-ViT zero-shot text-conditioned bounding box model
        """
        super().__init__()
        self.processor = OwlViTProcessor.from_pretrained("google/owlvit-base-patch32")
        self.processor.feature_extractor = OwlViTFeatureExtractor()
        self.model = OwlViTForObjectDetection.from_pretrained(
            "google/owlvit-base-patch32"
        )
        self.size = (768, 768)

    @property
    def device(self):
        return next(iter(self.parameters())).device

    def encode_texts(self, texts: List[List[str]]) -> OWLViTEncodings:
        try:
            processed = self.processor(text=texts, return_tensors="pt")
        except ValueError as e:
            raise ValueError(
                "Failed to encode texts, they are probably too long"
            ) from e
        return OWLViTEncodings(
            texts=texts,
            batch_encoding=processed.to(self.device),
        )

    def forward(
        self, images: Tensor.dims("NCHW"), encodings: OWLViTEncodings
    ) -> OWLViTPredictions:
        if images.shape[-2:] != self.size:
            images = resize(images, out_shape=self.size)

        images = TF.normalize(
            images,
            mean=self.processor.feature_extractor.image_mean,
            std=self.processor.feature_extractor.image_std,
        )

        inputs = BatchEncoding(
            data=dict(
                pixel_values=images,
                **encodings.repeat(images.shape[0]).batch_encoding,
            ),
            tensor_type="pt",
        )

        outputs = self.model(**inputs)
        target_sizes = torch.as_tensor([images.shape[-2:]]).repeat(len(images), 1)
        results = self.processor.post_process(
            outputs=outputs, target_sizes=target_sizes
        )

        boxes, scores, labels = zip(
            *[
                (result["boxes"], result["scores"], result["labels"])
                for result in results
            ]
        )

        return OWLViTPredictions(
            logits=outputs.logits,
            boxes=torch.stack(boxes),
            scores=torch.stack(scores),
            labels=torch.stack(labels),
            texts=encodings.texts,
        )


def test_owlvit():
    import requests
    from PIL import Image
    import torchvision.transforms.functional as TF
    import transformers

    image_url = "http://images.cocodataset.org/val2017/000000039769.jpg"
    images = (
        TF.to_tensor(Image.open(requests.get(image_url, stream=True).raw))[None]
        .cuda()
        .requires_grad_()
    )
    texts = [["a photo of a cat", "a photo of a dog"]]

    model = OWLViT().cuda()
    encodings = model.encode_texts(texts)

    with torch.enable_grad():
        # print(encodings.batch_encoding.device)
        predictions = model(images, encodings)
        predictions.scores.mean().backward()

    assert images.grad is not None

    images = resize(images, out_shape=(768, 768))
    images = TF.to_pil_image(images[0])

    reference_processor = transformers.OwlViTProcessor.from_pretrained(
        "google/owlvit-base-patch32"
    )
    inputs = reference_processor(text=texts, images=images, return_tensors="pt")

    reference_model = transformers.OwlViTForObjectDetection.from_pretrained(
        "google/owlvit-base-patch32"
    )
    outputs = reference_model(**inputs)
    target_sizes = torch.as_tensor([images.size[::-1]])
    results = reference_processor.post_process(
        outputs=outputs, target_sizes=target_sizes
    )

    boxes, scores = (
        results[0]["boxes"],
        results[0]["scores"],
    )

    assert torch.allclose(boxes[0], predictions.boxes[0, 0].cpu(), atol=10)
    assert torch.allclose(scores[0], predictions.scores[0, 0].cpu(), atol=1e-2)


def test_owlvit_too_long_text():
    import pytest

    text_prompts = [
        "photo of a warrior. trending on artstation. #photorealistic #artstation",
        "photo of a strong handsome warrior",
    ]
    with pytest.raises(ValueError):
        OWLViT().encode_texts(text_prompts)



# -*- coding: utf-8 -*-
"""
Useful utilities for testing the 2-D DTCWT with synthetic images
License: https://github.com/fbcotter/pytorch_wavelets/blob/master/LICENSE
Source: https://github.com/fbcotter/pytorch_wavelets/blob/31d6ac1b51b08f811a6a70eb7b3440f106009da0/pytorch_wavelets/dwt/lowlevel.py  # noqa
"""

import pywt
import torch
import numpy as np
import torch.nn.functional as F
from torch.autograd import Function


def sfb1d(lo, hi, g0, g1, mode='zero', dim=-1):
    """ 1D synthesis filter bank of an image tensor
    """
    C = lo.shape[1]
    d = dim % 4
    # If g0, g1 are not tensors, make them. If they are, then assume that they
    # are in the right order
    if not isinstance(g0, torch.Tensor):
        g0 = torch.tensor(np.copy(np.array(g0).ravel()),
                          dtype=torch.float, device=lo.device)
    if not isinstance(g1, torch.Tensor):
        g1 = torch.tensor(np.copy(np.array(g1).ravel()),
                          dtype=torch.float, device=lo.device)
    L = g0.numel()
    shape = [1, 1, 1, 1]
    shape[d] = L
    N = 2*lo.shape[d]
    # If g aren't in the right shape, make them so
    if g0.shape != tuple(shape):
        g0 = g0.reshape(*shape)
    if g1.shape != tuple(shape):
        g1 = g1.reshape(*shape)

    s = (2, 1) if d == 2 else (1, 2)
    g0 = torch.cat([g0]*C, dim=0)
    g1 = torch.cat([g1]*C, dim=0)
    if mode == 'per' or mode == 'periodization':
        y = F.conv_transpose2d(lo, g0, stride=s, groups=C) + \
            F.conv_transpose2d(hi, g1, stride=s, groups=C)
        if d == 2:
            y[:, :, :L-2] = y[:, :, :L-2] + y[:, :, N:N+L-2]
            y = y[:, :, :N]
        else:
            y[:, :, :, :L-2] = y[:, :, :, :L-2] + y[:, :, :, N:N+L-2]
            y = y[:, :, :, :N]
        y = roll(y, 1-L//2, dim=dim)
    else:
        if mode == 'zero' or mode == 'symmetric' or mode == 'reflect' or \
                mode == 'periodic':
            pad = (L-2, 0) if d == 2 else (0, L-2)
            y = F.conv_transpose2d(lo, g0, stride=s, padding=pad, groups=C) + \
                F.conv_transpose2d(hi, g1, stride=s, padding=pad, groups=C)
        else:
            raise ValueError('Unkown pad type: {}'.format(mode))

    return y


def _SFB2D(low, highs, g0_row, g1_row, g0_col, g1_col, mode):
    mode = int_to_mode(mode)

    lh, hl, hh = torch.unbind(highs, dim=2)
    lo = sfb1d(low, lh, g0_col, g1_col, mode=mode, dim=2)
    hi = sfb1d(hl, hh, g0_col, g1_col, mode=mode, dim=2)
    y = sfb1d(lo, hi, g0_row, g1_row, mode=mode, dim=3)

    return y


def roll(x, n, dim, make_even=False):
    if n < 0:
        n = x.shape[dim] + n

    if make_even and x.shape[dim] % 2 == 1:
        end = 1
    else:
        end = 0

    if dim == 0:
        return torch.cat((x[-n:], x[:-n+end]), dim=0)
    elif dim == 1:
        return torch.cat((x[:, -n:], x[:, :-n+end]), dim=1)
    elif dim == 2 or dim == -2:
        return torch.cat((x[:, :, -n:], x[:, :, :-n+end]), dim=2)
    elif dim == 3 or dim == -1:
        return torch.cat((x[:, :, :, -n:], x[:, :, :, :-n+end]), dim=3)


def int_to_mode(mode):
    if mode == 0:
        return 'zero'
    elif mode == 1:
        return 'symmetric'
    elif mode == 2:
        return 'periodization'
    elif mode == 3:
        return 'constant'
    elif mode == 4:
        return 'reflect'
    elif mode == 5:
        return 'replicate'
    elif mode == 6:
        return 'periodic'
    else:
        raise ValueError('Unkown pad type: {}'.format(mode))


def prep_filt_sfb2d(g0_col, g1_col, g0_row=None, g1_row=None, device=None):
    """
    Prepares the filters to be of the right form for the sfb2d function.  In
    particular, makes the tensors the right shape. It does not mirror image them
    as as sfb2d uses conv2d_transpose which acts like normal convolution.
    Inputs:
        g0_col (array-like): low pass column filter bank
        g1_col (array-like): high pass column filter bank
        g0_row (array-like): low pass row filter bank. If none, will assume the
            same as column filter
        g1_row (array-like): high pass row filter bank. If none, will assume the
            same as column filter
        device: which device to put the tensors on to
    Returns:
        (g0_col, g1_col, g0_row, g1_row)
    """
    g0_col, g1_col = prep_filt_sfb1d(g0_col, g1_col, device)
    if g0_row is None:
        g0_row, g1_row = g0_col, g1_col
    else:
        g0_row, g1_row = prep_filt_sfb1d(g0_row, g1_row, device)

    g0_col = g0_col.reshape((1, 1, -1, 1))
    g1_col = g1_col.reshape((1, 1, -1, 1))
    g0_row = g0_row.reshape((1, 1, 1, -1))
    g1_row = g1_row.reshape((1, 1, 1, -1))

    return g0_col, g1_col, g0_row, g1_row


def prep_filt_sfb1d(g0, g1, device=None):
    """
    Prepares the filters to be of the right form for the sfb1d function. In
    particular, makes the tensors the right shape. It does not mirror image them
    as as sfb2d uses conv2d_transpose which acts like normal convolution.
    Inputs:
        g0 (array-like): low pass filter bank
        g1 (array-like): high pass filter bank
        device: which device to put the tensors on to
    Returns:
        (g0, g1)
    """
    g0 = np.array(g0).ravel()
    g1 = np.array(g1).ravel()
    t = torch.get_default_dtype()
    g0 = torch.tensor(g0, device=device, dtype=t).reshape((1, 1, -1))
    g1 = torch.tensor(g1, device=device, dtype=t).reshape((1, 1, -1))

    return g0, g1


def mode_to_int(mode):
    if mode == 'zero':
        return 0
    elif mode == 'symmetric':
        return 1
    elif mode == 'per' or mode == 'periodization':
        return 2
    elif mode == 'constant':
        return 3
    elif mode == 'reflect':
        return 4
    elif mode == 'replicate':
        return 5
    elif mode == 'periodic':
        return 6
    else:
        raise ValueError('Unkown pad type: {}'.format(mode))


def afb1d(x, h0, h1, mode='zero', dim=-1):
    """ 1D analysis filter bank (along one dimension only) of an image
    Inputs:
        x (tensor): 4D input with the last two dimensions the spatial input
        h0 (tensor): 4D input for the lowpass filter. Should have shape (1, 1,
            h, 1) or (1, 1, 1, w)
        h1 (tensor): 4D input for the highpass filter. Should have shape (1, 1,
            h, 1) or (1, 1, 1, w)
        mode (str): padding method
        dim (int) - dimension of filtering. d=2 is for a vertical filter (called
            column filtering but filters across the rows). d=3 is for a
            horizontal filter, (called row filtering but filters across the
            columns).
    Returns:
        lohi: lowpass and highpass subbands concatenated along the channel
            dimension
    """
    C = x.shape[1]
    # Convert the dim to positive
    d = dim % 4
    s = (2, 1) if d == 2 else (1, 2)
    N = x.shape[d]
    # If h0, h1 are not tensors, make them. If they are, then assume that they
    # are in the right order
    if not isinstance(h0, torch.Tensor):
        h0 = torch.tensor(np.copy(np.array(h0).ravel()[::-1]),
                          dtype=torch.float, device=x.device)
    if not isinstance(h1, torch.Tensor):
        h1 = torch.tensor(np.copy(np.array(h1).ravel()[::-1]),
                          dtype=torch.float, device=x.device)
    L = h0.numel()
    L2 = L // 2
    shape = [1, 1, 1, 1]
    shape[d] = L
    # If h aren't in the right shape, make them so
    if h0.shape != tuple(shape):
        h0 = h0.reshape(*shape)
    if h1.shape != tuple(shape):
        h1 = h1.reshape(*shape)
    h = torch.cat([h0, h1] * C, dim=0)

    if mode == 'per' or mode == 'periodization':
        if x.shape[dim] % 2 == 1:
            if d == 2:
                x = torch.cat((x, x[:, :, -1:]), dim=2)
            else:
                x = torch.cat((x, x[:, :, :, -1:]), dim=3)
            N += 1
        x = roll(x, -L2, dim=d)
        pad = (L-1, 0) if d == 2 else (0, L-1)
        lohi = F.conv2d(x, h, padding=pad, stride=s, groups=C)
        N2 = N//2
        if d == 2:
            lohi[:, :, :L2] = lohi[:, :, :L2] + lohi[:, :, N2:N2+L2]
            lohi = lohi[:, :, :N2]
        else:
            lohi[:, :, :, :L2] = lohi[:, :, :, :L2] + lohi[:, :, :, N2:N2+L2]
            lohi = lohi[:, :, :, :N2]
    else:
        # Calculate the pad size
        outsize = pywt.dwt_coeff_len(N, L, mode=mode)
        p = 2 * (outsize - 1) - N + L
        if mode == 'zero':
            # Sadly, pytorch only allows for same padding before and after, if
            # we need to do more padding after for odd length signals, have to
            # prepad
            if p % 2 == 1:
                pad = (0, 0, 0, 1) if d == 2 else (0, 1, 0, 0)
                x = F.pad(x, pad)
            pad = (p//2, 0) if d == 2 else (0, p//2)
            # Calculate the high and lowpass
            lohi = F.conv2d(x, h, padding=pad, stride=s, groups=C)
        elif mode == 'symmetric' or mode == 'reflect' or mode == 'periodic':
            pad = (0, 0, p//2, (p+1)//2) if d == 2 else (p//2, (p+1)//2, 0, 0)
            x = mypad(x, pad=pad, mode=mode)
            lohi = F.conv2d(x, h, stride=s, groups=C)
        else:
            raise ValueError('Unkown pad type: {}'.format(mode))

    return lohi


def mypad(x, pad, mode='constant', value=0):
    """ Function to do numpy like padding on tensors. Only works for 2-D
    padding.
    Inputs:
        x (tensor): tensor to pad
        pad (tuple): tuple of (left, right, top, bottom) pad sizes
        mode (str): 'symmetric', 'wrap', 'constant, 'reflect', 'replicate', or
            'zero'. The padding technique.
    """
    if mode == 'symmetric':
        # Vertical only
        if pad[0] == 0 and pad[1] == 0:
            m1, m2 = pad[2], pad[3]
            l = x.shape[-2]  # noqa
            xe = reflect(np.arange(-m1, l+m2, dtype='int32'), -0.5, l-0.5)
            return x[:, :, xe]
        # horizontal only
        elif pad[2] == 0 and pad[3] == 0:
            m1, m2 = pad[0], pad[1]
            l = x.shape[-1]  # noqa
            xe = reflect(np.arange(-m1, l+m2, dtype='int32'), -0.5, l-0.5)
            return x[:, :, :, xe]
        # Both
        else:
            m1, m2 = pad[0], pad[1]
            l1 = x.shape[-1]
            xe_row = reflect(np.arange(-m1, l1+m2, dtype='int32'), -0.5, l1-0.5)
            m1, m2 = pad[2], pad[3]
            l2 = x.shape[-2]
            xe_col = reflect(np.arange(-m1, l2+m2, dtype='int32'), -0.5, l2-0.5)
            i = np.outer(xe_col, np.ones(xe_row.shape[0]))
            j = np.outer(np.ones(xe_col.shape[0]), xe_row)
            return x[:, :, i, j]
    elif mode == 'periodic':
        # Vertical only
        if pad[0] == 0 and pad[1] == 0:
            xe = np.arange(x.shape[-2])
            xe = np.pad(xe, (pad[2], pad[3]), mode='wrap')
            return x[:, :, xe]
        # Horizontal only
        elif pad[2] == 0 and pad[3] == 0:
            xe = np.arange(x.shape[-1])
            xe = np.pad(xe, (pad[0], pad[1]), mode='wrap')
            return x[:, :, :, xe]
        # Both
        else:
            xe_col = np.arange(x.shape[-2])
            xe_col = np.pad(xe_col, (pad[2], pad[3]), mode='wrap')
            xe_row = np.arange(x.shape[-1])
            xe_row = np.pad(xe_row, (pad[0], pad[1]), mode='wrap')
            i = np.outer(xe_col, np.ones(xe_row.shape[0]))
            j = np.outer(np.ones(xe_col.shape[0]), xe_row)
            return x[:, :, i, j]

    elif mode == 'constant' or mode == 'reflect' or mode == 'replicate':
        return F.pad(x, pad, mode, value)
    elif mode == 'zero':
        return F.pad(x, pad)
    else:
        raise ValueError('Unkown pad type: {}'.format(mode))


def reflect(x, minx, maxx):
    """Reflect the values in matrix *x* about the scalar values *minx* and
    *maxx*.  Hence a vector *x* containing a long linearly increasing series is
    converted into a waveform which ramps linearly up and down between *minx*
    and *maxx*.  If *x* contains integers and *minx* and *maxx* are (integers +
    0.5), the ramps will have repeated max and min samples.
    .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
    .. codeauthor:: Nick Kingsbury, Cambridge University, January 1999.
    """
    x = np.asanyarray(x)
    rng = maxx - minx
    rng_by_2 = 2 * rng
    mod = np.fmod(x - minx, rng_by_2)
    normed_mod = np.where(mod < 0, mod + rng_by_2, mod)
    out = np.where(normed_mod >= rng, rng_by_2 - normed_mod, normed_mod) + minx
    return np.array(out, dtype=x.dtype)


class SFB2D(Function):
    """ Does a single level 2d wavelet decomposition of an input. Does separate
    row and column filtering by two calls to
    :py:func:`pytorch_wavelets.dwt.lowlevel.afb1d`
    Needs to have the tensors in the right form. Because this function defines
    its own backward pass, saves on memory by not having to save the input
    tensors.
    Inputs:
        x (torch.Tensor): Input to decompose
        h0_row: row lowpass
        h1_row: row highpass
        h0_col: col lowpass
        h1_col: col highpass
        mode (int): use mode_to_int to get the int code here
    We encode the mode as an integer rather than a string as gradcheck causes an
    error when a string is provided.
    Returns:
        y: Tensor of shape (N, C*4, H, W)
    """
    @staticmethod
    def forward(ctx, low, highs, g0_row, g1_row, g0_col, g1_col, mode):
        mode = int_to_mode(mode)
        ctx.mode = mode
        ctx.save_for_backward(g0_row, g1_row, g0_col, g1_col)

        lh, hl, hh = torch.unbind(highs, dim=2)
        lo = sfb1d(low, lh, g0_col, g1_col, mode=mode, dim=2)
        hi = sfb1d(hl, hh, g0_col, g1_col, mode=mode, dim=2)
        y = sfb1d(lo, hi, g0_row, g1_row, mode=mode, dim=3)
        return y

    @staticmethod
    def backward(ctx, dy):
        dlow, dhigh = None, None
        if ctx.needs_input_grad[0]:
            mode = ctx.mode
            g0_row, g1_row, g0_col, g1_col = ctx.saved_tensors
            dx = afb1d(dy, g0_row, g1_row, mode=mode, dim=3)
            dx = afb1d(dx, g0_col, g1_col, mode=mode, dim=2)
            s = dx.shape
            dx = dx.reshape(s[0], -1, 4, s[-2], s[-1])
            dlow = dx[:, :, 0].contiguous()
            dhigh = dx[:, :, 1:].contiguous()
        return dlow, dhigh, None, None, None, None, None



model_urls = {
    "yfcc_2": "https://s3.eu-central-1.wasabisys.com/nextml-model-data/velocity-diffusion/yfcc_2.pth",
    "yfcc_1": "https://s3.eu-central-1.wasabisys.com/nextml-model-data/velocity-diffusion/yfcc_1.pth",
    "cc12m_1": "https://s3.eu-central-1.wasabisys.com/nextml-model-data/velocity-diffusion/cc12m_1.pth",
    "cc12m_1_cfg": "https://s3.eu-central-1.wasabisys.com/nextml-model-data/velocity-diffusion/cc12m_1_cfg.pth",
    "wikiart": "https://the-eye.eu/public/AI/models/v-diffusion/wikiart_256.pth",
}



import torch
import torch.nn as nn
import torch.nn.functional as F

from .mini_vit import mViT


class UpSampleBN(nn.Module):
    def __init__(self, skip_input, output_features):
        super(UpSampleBN, self).__init__()

        self._net = nn.Sequential(
            nn.Conv2d(skip_input, output_features, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(output_features),
            nn.LeakyReLU(),
            nn.Conv2d(
                output_features, output_features, kernel_size=3, stride=1, padding=1
            ),
            nn.BatchNorm2d(output_features),
            nn.LeakyReLU(),
        )

    def forward(self, x, concat_with):
        up_x = F.interpolate(
            x,
            size=[concat_with.size(2), concat_with.size(3)],
            mode="bilinear",
            align_corners=True,
        )
        f = torch.cat([up_x, concat_with], dim=1)
        return self._net(f)


class DecoderBN(nn.Module):
    def __init__(self, num_features=2048, num_classes=1, bottleneck_features=2048):
        super(DecoderBN, self).__init__()
        features = int(num_features)

        self.conv2 = nn.Conv2d(
            bottleneck_features, features, kernel_size=1, stride=1, padding=1
        )

        self.up1 = UpSampleBN(
            skip_input=features // 1 + 112 + 64, output_features=features // 2
        )
        self.up2 = UpSampleBN(
            skip_input=features // 2 + 40 + 24, output_features=features // 4
        )
        self.up3 = UpSampleBN(
            skip_input=features // 4 + 24 + 16, output_features=features // 8
        )
        self.up4 = UpSampleBN(
            skip_input=features // 8 + 16 + 8, output_features=features // 16
        )

        #         self.up5 = UpSample(skip_input=features // 16 + 3, output_features=features//16)
        self.conv3 = nn.Conv2d(
            features // 16, num_classes, kernel_size=3, stride=1, padding=1
        )
        # self.act_out = nn.Softmax(dim=1) if output_activation == 'softmax' else nn.Identity()

    def forward(self, features):
        x_block0, x_block1, x_block2, x_block3, x_block4 = (
            features[4],
            features[5],
            features[6],
            features[8],
            features[11],
        )

        x_d0 = self.conv2(x_block4)

        x_d1 = self.up1(x_d0, x_block3)
        x_d2 = self.up2(x_d1, x_block2)
        x_d3 = self.up3(x_d2, x_block1)
        x_d4 = self.up4(x_d3, x_block0)
        #         x_d5 = self.up5(x_d4, features[0])
        out = self.conv3(x_d4)
        # out = self.act_out(out)
        # if with_features:
        #     return out, features[-1]
        # elif with_intermediate:
        #     return out, [x_block0, x_block1, x_block2, x_block3, x_block4, x_d1, x_d2, x_d3, x_d4]
        return out


class Encoder(nn.Module):
    def __init__(self, backend):
        super(Encoder, self).__init__()
        self.original_model = backend

    def forward(self, x):
        features = [x]
        for k, v in self.original_model._modules.items():
            if k == "blocks":
                for ki, vi in v._modules.items():
                    features.append(vi(features[-1]))
            else:
                features.append(v(features[-1]))
        return features


class UnetAdaptiveBins(nn.Module):
    def __init__(self, backend, n_bins=100, min_val=0.1, max_val=10, norm="linear"):
        super(UnetAdaptiveBins, self).__init__()
        self.num_classes = n_bins
        self.min_val = min_val
        self.max_val = max_val
        self.encoder = Encoder(backend)
        self.adaptive_bins_layer = mViT(
            128,
            n_query_channels=128,
            patch_size=16,
            dim_out=n_bins,
            embedding_dim=128,
            norm=norm,
        )

        self.decoder = DecoderBN(num_classes=128)
        self.conv_out = nn.Sequential(
            nn.Conv2d(128, n_bins, kernel_size=1, stride=1, padding=0),
            nn.Softmax(dim=1),
        )

    def forward(self, x, **kwargs):
        unet_out = self.decoder(self.encoder(x), **kwargs)
        bin_widths_normed, range_attention_maps = self.adaptive_bins_layer(unet_out)
        out = self.conv_out(range_attention_maps)

        # Post process
        # n, c, h, w = out.shape
        # hist = torch.sum(out.view(n, c, h * w), dim=2) / (h * w)  # not used for training

        bin_widths = (
            self.max_val - self.min_val
        ) * bin_widths_normed  # .shape = N, dim_out
        bin_widths = nn.functional.pad(
            bin_widths, (1, 0), mode="constant", value=self.min_val
        )
        bin_edges = torch.cumsum(bin_widths, dim=1)

        centers = 0.5 * (bin_edges[:, :-1] + bin_edges[:, 1:])
        n, dout = centers.size()
        centers = centers.view(n, dout, 1, 1)

        pred = torch.sum(out * centers, dim=1, keepdim=True)

        return bin_edges, pred

    def get_1x_lr_params(self):  # lr/10 learning rate
        return self.encoder.parameters()

    def get_10x_lr_params(self):  # lr learning rate
        modules = [self.decoder, self.adaptive_bins_layer, self.conv_out]
        for m in modules:
            yield from m.parameters()

    @classmethod
    def build(cls, n_bins, **kwargs):
        basemodel_name = "tf_efficientnet_b5_ap"

        print("Loading base model ()...".format(basemodel_name), end="")
        basemodel = torch.hub.load(
            "rwightman/gen-efficientnet-pytorch", basemodel_name, pretrained=True
        )
        print("Done.")

        # Remove last layer
        print("Removing last two layers (global_pool & classifier).")
        basemodel.global_pool = nn.Identity()
        basemodel.classifier = nn.Identity()

        # Building Encoder-Decoder model
        print("Building Encoder-Decoder model..", end="")
        m = cls(basemodel, n_bins=n_bins, **kwargs)
        print("Done.")
        return m



import torch
from torch.nn import functional as F
from torchvision.transforms import functional as TF

from .interface import TransformInterface


class ClampWithGradFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, min=0, max=1):
        ctx.min = min
        ctx.max = max
        ctx.save_for_backward(input)
        return input.clamp(min, max)

    @staticmethod
    def backward(ctx, grad_in):
        (input,) = ctx.saved_tensors
        return (
            grad_in * (grad_in * (input - input.clamp(ctx.min, ctx.max)) >= 0),
            None,
            None,
        )


def clamp_with_grad(tensor, min=0.0, max=1.0):
    return ClampWithGradFunction.apply(tensor, min, max)


class ClampWithGrad(TransformInterface):
    def __init__(self, min=0, max=1):
        super().__init__()
        self.min = min
        self.max = max

    def encode(self, tensor):
        return clamp_with_grad(tensor, self.min, self.max)

    def decode(self, tensor):
        return tensor



from torch import nn
import torch.nn.functional as F

from perceptor.drawers.interface import DrawingInterface
from .compression import compress_jpeg
from .decompression import decompress_jpeg


class JPEG(DrawingInterface):
    def __init__(self, init_images, factor=1):
        super().__init__()
        self.shape = init_images.shape
        height, width, *_ = init_images.shape[-2:]
        self.compress_jpeg = compress_jpeg(factor=factor)
        self.decompress_jpeg = decompress_jpeg(height, width, factor=factor)
        self.ycbcr = nn.ParameterList(
            [nn.Parameter(parameter) for parameter in self.encode(init_images)]
        )

    def synthesize(self, _=None):
        return self.decode(self.ycbcr)

    def encode(self, image):
        return self.compress_jpeg(
            F.interpolate(image, size=self.shape[-2:], mode="bilinear")
        )

    def decode(self, ycbcr):
        return self.decompress_jpeg(*ycbcr)



import torch
import lantern


class Conditioning(torch.nn.Module):
    def __init__(self, encodings):
        super().__init__()
        self.encodings = torch.nn.Parameter(encodings, requires_grad=False)

    @property
    def device(self):
        return self.encodings.device

    def __neg__(self):
        return Conditioning(-self.encodings)



from .interface import LossInterface


class SphericalDistance(LossInterface):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, images_a, images_b):
        spherical_distance = (
            (
                self.model.encode_images(images_a)[:, None]
                - self.model.encode_images(images_b)[None, :]
            )
            .norm(dim=2)
            .div(2)
            .arcsin()
            .square()
            .mul(2)
        )
        return spherical_distance.mean()



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""Converting legacy network pickle into the new format."""

import io
import click
import pickle
import re
import copy
import numpy as np
import torch
import io

from . import misc, dnnlib, torch_utils

# ----------------------------------------------------------------------------


def load_network_pkl(f, force_fp16=False):
    data = _LegacyUnpickler(f).load()

    # Legacy TensorFlow pickle => convert.
    if (
        isinstance(data, tuple)
        and len(data) == 3
        and all(isinstance(net, _TFNetworkStub) for net in data)
    ):
        tf_G, tf_D, tf_Gs = data
        G = convert_tf_generator(tf_G)
        D = convert_tf_discriminator(tf_D)
        G_ema = convert_tf_generator(tf_Gs)
        data = dict(G=G, D=D, G_ema=G_ema)

    # Add missing fields.
    if "training_set_kwargs" not in data:
        data["training_set_kwargs"] = None
    if "augment_pipe" not in data:
        data["augment_pipe"] = None

    # Validate contents.
    assert isinstance(data["G"], torch.nn.Module)
    assert isinstance(data["G_ema"], torch.nn.Module)
    # assert isinstance(data['D'], torch.nn.Module)
    # assert isinstance(data['training_set_kwargs'], (dict, type(None)))
    # assert isinstance(data['augment_pipe'], (torch.nn.Module, type(None)))

    # Force FP16.
    if force_fp16:
        for key in ["G", "D", "G_ema"]:
            old = data[key]
            kwargs = copy.deepcopy(old.init_kwargs)
            fp16_kwargs = kwargs.get("synthesis_kwargs", kwargs)
            fp16_kwargs.num_fp16_res = 4
            fp16_kwargs.conv_clamp = 256
            if kwargs != old.init_kwargs:
                new = type(old)(**kwargs).eval().requires_grad_(False)
                misc.copy_params_and_buffers(old, new, require_all=True)
                data[key] = new
    return data


# ----------------------------------------------------------------------------


class _TFNetworkStub(dnnlib.EasyDict):
    pass


class _LegacyUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        if module == "dnnlib.tflib.network" and name == "Network":
            return _TFNetworkStub
        if module == "torch.storage" and name == "_load_from_bytes":
            return lambda b: torch.load(io.BytesIO(b), map_location="cpu")
        if module == "torch.storage" and name == "_load_from_bytes":
            return lambda b: torch.load(io.BytesIO(b), map_location="cpu")
        if module == "torch_utils":
            return getattr(torch_utils, name)
        if module == "torch_utils.persistence":
            return getattr(torch_utils.persistence, name)
        print(module, name)
        return super().find_class(module, name)


# ----------------------------------------------------------------------------


def _collect_tf_params(tf_net):
    # pylint: disable=protected-access
    tf_params = dict()

    def recurse(prefix, tf_net):
        for name, value in tf_net.variables:
            tf_params[prefix + name] = value
        for name, comp in tf_net.components.items():
            recurse(prefix + name + "/", comp)

    recurse("", tf_net)
    return tf_params


# ----------------------------------------------------------------------------


def _populate_module_params(module, *patterns):
    for name, tensor in misc.named_params_and_buffers(module):
        found = False
        value = None
        for pattern, value_fn in zip(patterns[0::2], patterns[1::2]):
            match = re.fullmatch(pattern, name)
            if match:
                found = True
                if value_fn is not None:
                    value = value_fn(*match.groups())
                break
        try:
            assert found
            if value is not None:
                tensor.copy_(torch.from_numpy(np.array(value)))
        except:
            print(name, list(tensor.shape))
            raise


# ----------------------------------------------------------------------------


def convert_tf_generator(tf_G):
    if tf_G.version < 4:
        raise ValueError("TensorFlow pickle version too low")

    # Collect kwargs.
    tf_kwargs = tf_G.static_kwargs
    known_kwargs = set()

    def kwarg(tf_name, default=None, none=None):
        known_kwargs.add(tf_name)
        val = tf_kwargs.get(tf_name, default)
        return val if val is not None else none

    # Convert kwargs.
    from training import networks_stylegan2

    network_class = networks_stylegan2.Generator
    kwargs = dnnlib.EasyDict(
        z_dim=kwarg("latent_size", 512),
        c_dim=kwarg("label_size", 0),
        w_dim=kwarg("dlatent_size", 512),
        img_resolution=kwarg("resolution", 1024),
        img_channels=kwarg("num_channels", 3),
        channel_base=kwarg("fmap_base", 16384) * 2,
        channel_max=kwarg("fmap_max", 512),
        num_fp16_res=kwarg("num_fp16_res", 0),
        conv_clamp=kwarg("conv_clamp", None),
        architecture=kwarg("architecture", "skip"),
        resample_filter=kwarg("resample_kernel", [1, 3, 3, 1]),
        use_noise=kwarg("use_noise", True),
        activation=kwarg("nonlinearity", "lrelu"),
        mapping_kwargs=dnnlib.EasyDict(
            num_layers=kwarg("mapping_layers", 8),
            embed_features=kwarg("label_fmaps", None),
            layer_features=kwarg("mapping_fmaps", None),
            activation=kwarg("mapping_nonlinearity", "lrelu"),
            lr_multiplier=kwarg("mapping_lrmul", 0.01),
            w_avg_beta=kwarg("w_avg_beta", 0.995, none=1),
        ),
    )

    # Check for unknown kwargs.
    kwarg("truncation_psi")
    kwarg("truncation_cutoff")
    kwarg("style_mixing_prob")
    kwarg("structure")
    kwarg("conditioning")
    kwarg("fused_modconv")
    unknown_kwargs = list(set(tf_kwargs.keys()) - known_kwargs)
    if len(unknown_kwargs) > 0:
        raise ValueError("Unknown TensorFlow kwarg", unknown_kwargs[0])

    # Collect params.
    tf_params = _collect_tf_params(tf_G)
    for name, value in list(tf_params.items()):
        match = re.fullmatch(r"ToRGB_lod(\d+)/(.*)", name)
        if match:
            r = kwargs.img_resolution // (2 ** int(match.group(1)))
            tf_params[f"{r}x{r}/ToRGB/{match.group(2)}"] = value
            kwargs.synthesis.kwargs.architecture = "orig"
    # for name, value in tf_params.items(): print(f'{name:<50s}{list(value.shape)}')

    # Convert params.
    G = network_class(**kwargs).eval().requires_grad_(False)
    # pylint: disable=unnecessary-lambda
    # pylint: disable=f-string-without-interpolation
    _populate_module_params(
        G,
        r"mapping\.w_avg",
        lambda: tf_params[f"dlatent_avg"],
        r"mapping\.embed\.weight",
        lambda: tf_params[f"mapping/LabelEmbed/weight"].transpose(),
        r"mapping\.embed\.bias",
        lambda: tf_params[f"mapping/LabelEmbed/bias"],
        r"mapping\.fc(\d+)\.weight",
        lambda i: tf_params[f"mapping/Dense{i}/weight"].transpose(),
        r"mapping\.fc(\d+)\.bias",
        lambda i: tf_params[f"mapping/Dense{i}/bias"],
        r"synthesis\.b4\.const",
        lambda: tf_params[f"synthesis/4x4/Const/const"][0],
        r"synthesis\.b4\.conv1\.weight",
        lambda: tf_params[f"synthesis/4x4/Conv/weight"].transpose(3, 2, 0, 1),
        r"synthesis\.b4\.conv1\.bias",
        lambda: tf_params[f"synthesis/4x4/Conv/bias"],
        r"synthesis\.b4\.conv1\.noise_const",
        lambda: tf_params[f"synthesis/noise0"][0, 0],
        r"synthesis\.b4\.conv1\.noise_strength",
        lambda: tf_params[f"synthesis/4x4/Conv/noise_strength"],
        r"synthesis\.b4\.conv1\.affine\.weight",
        lambda: tf_params[f"synthesis/4x4/Conv/mod_weight"].transpose(),
        r"synthesis\.b4\.conv1\.affine\.bias",
        lambda: tf_params[f"synthesis/4x4/Conv/mod_bias"] + 1,
        r"synthesis\.b(\d+)\.conv0\.weight",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv0_up/weight"][::-1, ::-1].transpose(
            3, 2, 0, 1
        ),
        r"synthesis\.b(\d+)\.conv0\.bias",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv0_up/bias"],
        r"synthesis\.b(\d+)\.conv0\.noise_const",
        lambda r: tf_params[f"synthesis/noise{int(np.log2(int(r)))*2-5}"][0, 0],
        r"synthesis\.b(\d+)\.conv0\.noise_strength",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv0_up/noise_strength"],
        r"synthesis\.b(\d+)\.conv0\.affine\.weight",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv0_up/mod_weight"].transpose(),
        r"synthesis\.b(\d+)\.conv0\.affine\.bias",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv0_up/mod_bias"] + 1,
        r"synthesis\.b(\d+)\.conv1\.weight",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv1/weight"].transpose(3, 2, 0, 1),
        r"synthesis\.b(\d+)\.conv1\.bias",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv1/bias"],
        r"synthesis\.b(\d+)\.conv1\.noise_const",
        lambda r: tf_params[f"synthesis/noise{int(np.log2(int(r)))*2-4}"][0, 0],
        r"synthesis\.b(\d+)\.conv1\.noise_strength",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv1/noise_strength"],
        r"synthesis\.b(\d+)\.conv1\.affine\.weight",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv1/mod_weight"].transpose(),
        r"synthesis\.b(\d+)\.conv1\.affine\.bias",
        lambda r: tf_params[f"synthesis/{r}x{r}/Conv1/mod_bias"] + 1,
        r"synthesis\.b(\d+)\.torgb\.weight",
        lambda r: tf_params[f"synthesis/{r}x{r}/ToRGB/weight"].transpose(3, 2, 0, 1),
        r"synthesis\.b(\d+)\.torgb\.bias",
        lambda r: tf_params[f"synthesis/{r}x{r}/ToRGB/bias"],
        r"synthesis\.b(\d+)\.torgb\.affine\.weight",
        lambda r: tf_params[f"synthesis/{r}x{r}/ToRGB/mod_weight"].transpose(),
        r"synthesis\.b(\d+)\.torgb\.affine\.bias",
        lambda r: tf_params[f"synthesis/{r}x{r}/ToRGB/mod_bias"] + 1,
        r"synthesis\.b(\d+)\.skip\.weight",
        lambda r: tf_params[f"synthesis/{r}x{r}/Skip/weight"][::-1, ::-1].transpose(
            3, 2, 0, 1
        ),
        r".*\.resample_filter",
        None,
        r".*\.act_filter",
        None,
    )
    return G


# ----------------------------------------------------------------------------


def convert_tf_discriminator(tf_D):
    if tf_D.version < 4:
        raise ValueError("TensorFlow pickle version too low")

    # Collect kwargs.
    tf_kwargs = tf_D.static_kwargs
    known_kwargs = set()

    def kwarg(tf_name, default=None):
        known_kwargs.add(tf_name)
        return tf_kwargs.get(tf_name, default)

    # Convert kwargs.
    kwargs = dnnlib.EasyDict(
        c_dim=kwarg("label_size", 0),
        img_resolution=kwarg("resolution", 1024),
        img_channels=kwarg("num_channels", 3),
        architecture=kwarg("architecture", "resnet"),
        channel_base=kwarg("fmap_base", 16384) * 2,
        channel_max=kwarg("fmap_max", 512),
        num_fp16_res=kwarg("num_fp16_res", 0),
        conv_clamp=kwarg("conv_clamp", None),
        cmap_dim=kwarg("mapping_fmaps", None),
        block_kwargs=dnnlib.EasyDict(
            activation=kwarg("nonlinearity", "lrelu"),
            resample_filter=kwarg("resample_kernel", [1, 3, 3, 1]),
            freeze_layers=kwarg("freeze_layers", 0),
        ),
        mapping_kwargs=dnnlib.EasyDict(
            num_layers=kwarg("mapping_layers", 0),
            embed_features=kwarg("mapping_fmaps", None),
            layer_features=kwarg("mapping_fmaps", None),
            activation=kwarg("nonlinearity", "lrelu"),
            lr_multiplier=kwarg("mapping_lrmul", 0.1),
        ),
        epilogue_kwargs=dnnlib.EasyDict(
            mbstd_group_size=kwarg("mbstd_group_size", None),
            mbstd_num_channels=kwarg("mbstd_num_features", 1),
            activation=kwarg("nonlinearity", "lrelu"),
        ),
    )

    # Check for unknown kwargs.
    kwarg("structure")
    kwarg("conditioning")
    unknown_kwargs = list(set(tf_kwargs.keys()) - known_kwargs)
    if len(unknown_kwargs) > 0:
        raise ValueError("Unknown TensorFlow kwarg", unknown_kwargs[0])

    # Collect params.
    tf_params = _collect_tf_params(tf_D)
    for name, value in list(tf_params.items()):
        match = re.fullmatch(r"FromRGB_lod(\d+)/(.*)", name)
        if match:
            r = kwargs.img_resolution // (2 ** int(match.group(1)))
            tf_params[f"{r}x{r}/FromRGB/{match.group(2)}"] = value
            kwargs.architecture = "orig"
    # for name, value in tf_params.items(): print(f'{name:<50s}{list(value.shape)}')

    # Convert params.
    from training import networks_stylegan2

    D = networks_stylegan2.Discriminator(**kwargs).eval().requires_grad_(False)
    # pylint: disable=unnecessary-lambda
    # pylint: disable=f-string-without-interpolation
    _populate_module_params(
        D,
        r"b(\d+)\.fromrgb\.weight",
        lambda r: tf_params[f"{r}x{r}/FromRGB/weight"].transpose(3, 2, 0, 1),
        r"b(\d+)\.fromrgb\.bias",
        lambda r: tf_params[f"{r}x{r}/FromRGB/bias"],
        r"b(\d+)\.conv(\d+)\.weight",
        lambda r, i: tf_params[
            f'{r}x{r}/Conv{i}{["","_down"][int(i)]}/weight'
        ].transpose(3, 2, 0, 1),
        r"b(\d+)\.conv(\d+)\.bias",
        lambda r, i: tf_params[f'{r}x{r}/Conv{i}{["","_down"][int(i)]}/bias'],
        r"b(\d+)\.skip\.weight",
        lambda r: tf_params[f"{r}x{r}/Skip/weight"].transpose(3, 2, 0, 1),
        r"mapping\.embed\.weight",
        lambda: tf_params[f"LabelEmbed/weight"].transpose(),
        r"mapping\.embed\.bias",
        lambda: tf_params[f"LabelEmbed/bias"],
        r"mapping\.fc(\d+)\.weight",
        lambda i: tf_params[f"Mapping{i}/weight"].transpose(),
        r"mapping\.fc(\d+)\.bias",
        lambda i: tf_params[f"Mapping{i}/bias"],
        r"b4\.conv\.weight",
        lambda: tf_params[f"4x4/Conv/weight"].transpose(3, 2, 0, 1),
        r"b4\.conv\.bias",
        lambda: tf_params[f"4x4/Conv/bias"],
        r"b4\.fc\.weight",
        lambda: tf_params[f"4x4/Dense0/weight"].transpose(),
        r"b4\.fc\.bias",
        lambda: tf_params[f"4x4/Dense0/bias"],
        r"b4\.out\.weight",
        lambda: tf_params[f"Output/weight"].transpose(),
        r"b4\.out\.bias",
        lambda: tf_params[f"Output/bias"],
        r".*\.resample_filter",
        None,
    )
    return D


# ----------------------------------------------------------------------------


@click.command()
@click.option("--source", help="Input pickle", required=True, metavar="PATH")
@click.option("--dest", help="Output pickle", required=True, metavar="PATH")
@click.option(
    "--force-fp16",
    help="Force the networks to use FP16",
    type=bool,
    default=False,
    metavar="BOOL",
    show_default=True,
)
def convert_network_pickle(source, dest, force_fp16):
    """Convert legacy network pickle into the native PyTorch format.
    The tool is able to load the main network configurations exported using the TensorFlow version of StyleGAN2 or StyleGAN2-ADA.
    It does not support e.g. StyleGAN2-ADA comparison methods, StyleGAN2 configs A-D, or StyleGAN1 networks.
    Example:
    \b
    python legacy.py \\
        --source=https://nvlabs-fi-cdn.nvidia.com/stylegan2/networks/stylegan2-cat-config-f.pkl \\
        --dest=stylegan2-cat-config-f.pkl
    """
    print(f'Loading "{source}"...')
    with dnnlib.util.open_url(source) as f:
        data = load_network_pkl(f, force_fp16=force_fp16)
    print(f'Saving "{dest}"...')
    with open(dest, "wb") as f:
        pickle.dump(data, f)
    print("Done.")


# ----------------------------------------------------------------------------

if __name__ == "__main__":
    convert_network_pickle()  # pylint: disable=no-value-for-parameter

# ----------------------------------------------------------------------------
